"""Classes for RESQML objects related to surfaces."""

__all__ = [
    "GridSkin",
    "generate_untorn_surface_for_layer_interface",
    "generate_torn_surface_for_layer_interface",
    "generate_torn_surface_for_x_section",
    "generate_untorn_surface_for_x_section",
    "point_is_within_cell",
    "create_column_face_mesh_and_surface",
    "find_intersections_of_trajectory_with_surface",
    "find_intersections_of_trajectory_with_layer_interface",
    "find_first_intersection_of_trajectory_with_surface",
    "find_first_intersection_of_trajectory_with_layer_interface",
    "find_first_intersection_of_trajectory_with_cell_surface",
    "find_intersection_of_trajectory_interval_with_column_face",
    "trajectory_grid_overlap",
    "populate_blocked_well_from_trajectory",
    "generate_surface_for_blocked_well_cells",
    "find_faces_to_represent_surface_staffa",
    "find_faces_to_represent_surface_regular",
    "find_faces_to_represent_surface_regular_optimised",
    "find_faces_to_represent_surface",
    "bisector_from_faces",
    "column_bisector_from_faces",
    "shadow_from_faces",
    "get_boundary",
    "_where_true",
    "_first_true",
    "intersect_numba",
]

from ._grid_skin import GridSkin
from ._grid_surface import (
    generate_untorn_surface_for_layer_interface,
    generate_torn_surface_for_layer_interface,
    generate_torn_surface_for_x_section,
    generate_untorn_surface_for_x_section,
    point_is_within_cell,
    create_column_face_mesh_and_surface,
)
from ._trajectory_intersects import (
    find_intersections_of_trajectory_with_surface,
    find_intersections_of_trajectory_with_layer_interface,
    find_first_intersection_of_trajectory_with_surface,
    find_first_intersection_of_trajectory_with_layer_interface,
    find_first_intersection_of_trajectory_with_cell_surface,
    find_intersection_of_trajectory_interval_with_column_face,
    trajectory_grid_overlap,
)
from ._blocked_well_populate import (
    populate_blocked_well_from_trajectory,
    generate_surface_for_blocked_well_cells,
)
from ._find_faces import (
    find_faces_to_represent_surface_staffa,
    find_faces_to_represent_surface_regular,
    find_faces_to_represent_surface_regular_optimised,
    find_faces_to_represent_surface,
    bisector_from_faces,
    column_bisector_from_faces,
    shadow_from_faces,
    get_boundary,
    _where_true,
    _first_true,
    intersect_numba,
)

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Class for RESQML Geologic Unit Feature organizational objects."""

import resqpy.olio.uuid as bu
import resqpy.organize
import resqpy.organize._utils as ou
from resqpy.olio.base import BaseResqpy


class GeologicUnitFeature(BaseResqpy):
    """Class for RESQML Geologic Unit Feature organizational objects."""

    resqml_type = "GeologicUnitFeature"
    feature_name = ou.alias_for_attribute("title")

    def __init__(self, parent_model, uuid = None, feature_name = None, originator = None, extra_metadata = None):
        """Initialises a geologic unit feature organisational object."""

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = feature_name,
                         originator = originator,
                         extra_metadata = extra_metadata)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this feature is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, GeologicUnitFeature):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        return self.feature_name == other.feature_name

    def create_xml(self, add_as_part = True, originator = None, reuse = True):
        """Creates a geologic unit feature organisational xml node from this geologic unit feature object."""
        if reuse and self.try_reuse():
            return self.root  # check for reusable (equivalent) object
        return super().create_xml(add_as_part = add_as_part, originator = originator)



"""Class for RESQML Generic Feature Interpretation objects."""

import resqpy.organize._utils as ou
import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class GenericInterpretation(BaseResqpy):
    """Class for RESQML Generic Feature Interpretation objects."""

    resqml_type = 'GenericFeatureInterpretation'
    valid_domains = ('depth', 'time', 'mixed')

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 feature_uuid = None,
                 domain = 'depth',
                 composition = None,
                 material_implacement = None,
                 geobody_shape = None,
                 originator = None,
                 extra_metadata = None):
        """Initialise a new geobody interpretation object, either from xml or explicitly."""

        assert domain in self.valid_domains

        self.uuid = None
        self.title = None
        self.feature_uuid = feature_uuid
        self.domain = domain
        self.has_occurred_during = (None, None)

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if not self.title and feature_uuid is not None:
            title = parent_model.title_for_root(parent_model.root_for_uuid(feature_uuid))

    def _load_from_xml(self):
        root_node = self.root
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        interp_feature_uuid = rqet.find_nested_tags_text(root_node, ['InterpretedFeature', 'UUID'])
        assert interp_feature_uuid is not None
        self.feature_uuid = bu.uuid_from_string(interp_feature_uuid)
        self.has_occurred_during = ou.extract_has_occurred_during(root_node)

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, GenericInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.title != other.title:
            return False
        if not bu.matching_uuids(self.feature_uuid, other.feature_uuid):
            return False
        if self.domain != other.domain:
            return False
        if not ou.equivalent_chrono_pairs(self.main_has_occurred_during, other.main_has_occurred_during):
            return False
        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        return True

    def create_xml(self, add_as_part = True, add_relationships = True, originator = None, reuse = True):
        """Creates a generic feature interpretation organisational xml node from a generic interpretation object."""

        # note: related feature node should be created first and referenced here
        assert self.feature_uuid is not None
        feature_root = self.model.root_for_uuid(self.feature_uuid)
        assert feature_root is not None

        if not self.title:
            self.title = rqet.find_nested_tags_text(feature_root, ['Citation', 'Title'])

        if reuse and self.try_reuse():
            return self.root

        gfi = super().create_xml(add_as_part = False, originator = originator)

        assert self.domain in self.valid_domains, 'illegal domain value for fault interpretation'
        dom_node = rqet.SubElement(gfi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(feature_root),
                                   self.feature_uuid,
                                   content_type = self.model.type_of_uuid(self.feature_uuid),
                                   root = gfi)

        ou.create_xml_has_occurred_during(self.model, gfi, self.has_occurred_during)

        if add_as_part:
            self.model.add_part('obj_GenericFeatureInterpretation', self.uuid, gfi)
            if add_relationships:
                self.model.create_reciprocal_relationship(gfi, 'destinationObject', feature_root, 'sourceObject')

        return gfi



"""HexaGrid class module."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.grid as grr
import resqpy.olio.volume as vol
import resqpy.property as rqp
import resqpy.unstructured
import resqpy.unstructured._unstructured_grid as rug


class HexaGrid(rug.UnstructuredGrid):
    """Class for unstructured grids where every cell is hexahedral (faces may be degenerate)."""

    def __init__(self,
                 parent_model,
                 uuid = None,
                 find_properties = True,
                 cache_geometry = False,
                 title = None,
                 originator = None,
                 extra_metadata = {}):
        """Creates a new resqpy HexaGrid object (RESQML UnstructuredGrid with cell shape hexahedral)

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           uuid (uuid.UUID, optional): if present, the new grid object is populated from the RESQML object
           find_properties (boolean, default True): if True and uuid is present, a
              grid property collection is instantiated as an attribute, holding properties for which
              this grid is the supporting representation
           cache_geometry (boolean, default False): if True and uuid is present, all the geometry arrays
              are loaded into attributes of the new grid object
           title (str, optional): citation title for new grid; ignored if uuid is present
           originator (str, optional): name of person creating the grid; defaults to login id;
              ignored if uuid is present
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid;
              ignored if uuid is present

        returns:
           a newly created HexaGrid object
        """

        super().__init__(parent_model = parent_model,
                         uuid = uuid,
                         find_properties = find_properties,
                         geometry_required = True,
                         cache_geometry = cache_geometry,
                         cell_shape = 'hexahedral',
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            assert grr.grid_flavour(self.root) == 'HexaGrid'
            self.check_hexahedral()

        self.grid_representation = 'HexaGrid'  #: flavour of grid; not much used

    @classmethod
    def from_unsplit_grid(cls,
                          parent_model,
                          grid_uuid,
                          inherit_properties = True,
                          title = None,
                          extra_metadata = {},
                          write_active = None):
        """Creates a new (unstructured) HexaGrid from an existing resqpy unsplit (IJK) Grid without K gaps.

        arguments:
           parent_model (model.Model object): the model which this grid is part of
           grid_uuid (uuid.UUID): the uuid of an IjkGridRepresentation from which the hexa grid will be created
           inherit_properties (boolean, default True): if True, properties will be created for the new grid
           title (str, optional): citation title for the new grid
           extra_metadata (dict, optional): dictionary of extra metadata items to add to the grid
           write_active (boolean, optional): if True (or None and inactive property is established) then an
              active cell property is created (in addition to any inherited properties)

        returns:
           a newly created HexaGrid object

        note:
           this method includes the writing of hdf5 data, creation of xml for the new grid and adding it as a part
        """

        import resqpy.grid as grr

        # establish existing IJK grid
        ijk_grid = grr.Grid(parent_model, uuid = grid_uuid, find_properties = inherit_properties)
        assert ijk_grid is not None
        assert not ijk_grid.has_split_coordinate_lines, 'IJK grid has split coordinate lines (faults)'
        assert not ijk_grid.k_gaps, 'IJK grid has K gaps'
        ijk_grid.cache_all_geometry_arrays()
        ijk_points = ijk_grid.points_ref(masked = False)
        if title is None:
            title = ijk_grid.title

        # make empty unstructured hexa grid
        hexa_grid = cls(parent_model, title = title, extra_metadata = extra_metadata)

        # derive hexa grid attributes from ijk grid
        hexa_grid.crs_uuid = ijk_grid.crs_uuid
        hexa_grid.set_cell_count(ijk_grid.cell_count())
        if ijk_grid.inactive is not None:
            hexa_grid.inactive = ijk_grid.inactive.reshape((hexa_grid.cell_count,))
            hexa_grid.all_inactive = np.all(hexa_grid.inactive)
            if hexa_grid.all_inactive:
                log.warning(f'all cells marked as inactive for unstructured hexa grid {hexa_grid.title}')
        else:
            hexa_grid.all_inactive = False

        # inherit points (nodes) in IJK grid order, ie. K cycling fastest, then I, then J
        hexa_grid.points_cached = ijk_points.reshape((-1, 3))

        # setup faces per cell
        # ordering of faces (in nodes per face): all K faces, then all J faces, then all I faces
        # within J faces, ordering is all of J- faces for J = 0 first, then increasing planes in J
        # similarly for I faces
        nk_plus_1 = ijk_grid.nk + 1
        nj_plus_1 = ijk_grid.nj + 1
        ni_plus_1 = ijk_grid.ni + 1
        k_face_count = nk_plus_1 * ijk_grid.nj * ijk_grid.ni
        j_face_count = ijk_grid.nk * nj_plus_1 * ijk_grid.ni
        i_face_count = ijk_grid.nk * ijk_grid.nj * ni_plus_1
        kj_face_count = k_face_count + j_face_count
        hexa_grid.face_count = k_face_count + j_face_count + i_face_count
        hexa_grid.faces_per_cell_cl = 6 * (1 + np.arange(hexa_grid.cell_count, dtype = int))  # 6 faces per cell
        hexa_grid.faces_per_cell = np.empty(6 * hexa_grid.cell_count, dtype = int)
        arange = np.arange(hexa_grid.cell_count, dtype = int)
        hexa_grid.faces_per_cell[0::6] = arange  # K- faces
        hexa_grid.faces_per_cell[1::6] = ijk_grid.nj * ijk_grid.ni + arange  # K+ faces
        nki = ijk_grid.nk * ijk_grid.ni
        nkj = ijk_grid.nk * ijk_grid.nj
        # todo: vectorise following for loop
        for cell in range(hexa_grid.cell_count):
            k, j, i = ijk_grid.denaturalized_cell_index(cell)
            j_minus_face = k_face_count + nki * j + ijk_grid.ni * k + i
            hexa_grid.faces_per_cell[6 * cell + 2] = j_minus_face  # J- face
            hexa_grid.faces_per_cell[6 * cell + 3] = j_minus_face + nki  # J+ face
            i_minus_face = kj_face_count + nkj * i + ijk_grid.nj * k + j
            hexa_grid.faces_per_cell[6 * cell + 4] = i_minus_face  # I- face
            hexa_grid.faces_per_cell[6 * cell + 5] = i_minus_face + nkj  # I+ face

        # setup nodes per face, clockwise when viewed from negative side of face if ijk handedness matches xyz handedness
        # ordering of nodes in points array is as for the IJK grid
        hexa_grid.node_count = hexa_grid.points_cached.shape[0]
        assert hexa_grid.node_count == (ijk_grid.nk + 1) * (ijk_grid.nj + 1) * (ijk_grid.ni + 1)
        hexa_grid.nodes_per_face_cl = 4 * (1 + np.arange(hexa_grid.face_count, dtype = int))  # 4 nodes per face
        hexa_grid.nodes_per_face = np.empty(4 * hexa_grid.face_count, dtype = int)
        # todo: vectorise for loops
        # K faces
        face_base = 0
        for k in range(nk_plus_1):
            for j in range(ijk_grid.nj):
                for i in range(ijk_grid.ni):
                    hexa_grid.nodes_per_face[face_base] = (k * nj_plus_1 + j) * ni_plus_1 + i  # ip 0, jp 0
                    hexa_grid.nodes_per_face[face_base + 1] = (k * nj_plus_1 + j + 1) * ni_plus_1 + i  # ip 0, jp 1
                    hexa_grid.nodes_per_face[face_base + 2] = (k * nj_plus_1 + j + 1) * ni_plus_1 + i + 1  # ip 1, jp 1
                    hexa_grid.nodes_per_face[face_base + 3] = (k * nj_plus_1 + j) * ni_plus_1 + i + 1  # ip 1, jp 0
                    face_base += 4
        # J faces
        assert face_base == 4 * k_face_count
        for j in range(nj_plus_1):
            for k in range(ijk_grid.nk):
                for i in range(ijk_grid.ni):
                    hexa_grid.nodes_per_face[face_base] = (k * nj_plus_1 + j) * ni_plus_1 + i  # ip 0, kp 0
                    hexa_grid.nodes_per_face[face_base + 1] = (k * nj_plus_1 + j) * ni_plus_1 + i + 1  # ip 1, kp 0
                    hexa_grid.nodes_per_face[face_base +
                                             2] = ((k + 1) * nj_plus_1 + j) * ni_plus_1 + i + 1  # ip 1, kp 1
                    hexa_grid.nodes_per_face[face_base + 3] = ((k + 1) * nj_plus_1 + j) * ni_plus_1 + i  # ip 0, kp 1
                    face_base += 4
        # I faces
        assert face_base == 4 * kj_face_count
        for i in range(ni_plus_1):
            for k in range(ijk_grid.nk):
                for j in range(ijk_grid.nj):
                    hexa_grid.nodes_per_face[face_base] = (k * nj_plus_1 + j) * ni_plus_1 + i  # jp 0, kp 0
                    hexa_grid.nodes_per_face[face_base + 1] = ((k + 1) * nj_plus_1 + j) * ni_plus_1 + i  # jp 0, kp 1
                    hexa_grid.nodes_per_face[face_base +
                                             2] = ((k + 1) * nj_plus_1 + j + 1) * ni_plus_1 + i  # jp 1, kp 1
                    hexa_grid.nodes_per_face[face_base + 3] = (k * nj_plus_1 + j + 1) * ni_plus_1 + i  # jp 1, kp 0
                    face_base += 4
        assert face_base == 4 * hexa_grid.face_count

        # set cell face is right handed
        # todo: check Energistics documents for meaning of cell face is right handed
        # here the assumption is clockwise ordering of nodes viewed from within cell means 'right handed'
        hexa_grid.cell_face_is_right_handed = np.zeros(6 * hexa_grid.cell_count,
                                                       dtype = bool)  # initially set to left handed
        # if IJK grid's ijk handedness matches the xyz handedness, then set +ve faces to right handed; else -ve faces
        if ijk_grid.off_handed():
            hexa_grid.cell_face_is_right_handed[0::2] = True  # negative faces are right handed
        else:
            hexa_grid.cell_face_is_right_handed[1::2] = True  # positive faces are right handed

        hexa_grid.write_hdf5(write_active = write_active)
        hexa_grid.create_xml(write_active = write_active)

        if inherit_properties:
            ijk_pc = ijk_grid.extract_property_collection()
            hexa_pc = rqp.PropertyCollection(support = hexa_grid)
            for part in ijk_pc.parts():
                count = ijk_pc.count_for_part(part)
                hexa_part_shape = (hexa_grid.cell_count,) if count == 1 else (hexa_grid.cell_count, count)
                hexa_pc.add_cached_array_to_imported_list(
                    ijk_pc.cached_part_array_ref(part).reshape(hexa_part_shape),
                    'inherited from grid ' + str(ijk_grid.title),
                    ijk_pc.citation_title_for_part(part),
                    discrete = not ijk_pc.continuous_for_part(part),
                    uom = ijk_pc.uom_for_part(part),
                    time_index = ijk_pc.time_index_for_part(part),
                    null_value = ijk_pc.null_value_for_part(part),
                    property_kind = ijk_pc.property_kind_for_part(part),
                    local_property_kind_uuid = ijk_pc.local_property_kind_uuid(part),
                    facet_type = ijk_pc.facet_type_for_part(part),
                    facet = ijk_pc.facet_for_part(part),
                    realization = ijk_pc.realization_for_part(part),
                    indexable_element = ijk_pc.indexable_for_part(part),
                    count = count,
                    const_value = ijk_pc.constant_value_for_part(part))
                # todo: patch min & max values if present in ijk part
                hexa_pc.write_hdf5_for_imported_list()
                hexa_pc.create_xml_for_imported_list_and_add_parts_to_model(
                    support_uuid = hexa_grid.uuid,
                    time_series_uuid = ijk_pc.time_series_uuid_for_part(part),
                    string_lookup_uuid = ijk_pc.string_lookup_uuid_for_part(part),
                    extra_metadata = ijk_pc.extra_metadata_for_part(part))

        return hexa_grid

    def check_hexahedral(self):
        """Checks that each cell has 6 faces and each face has 4 nodes.

        notes:
           currently only performs a cursory check, without checking nodes are shared;
           assumes that degenerate faces still have four nodes identified
        """

        assert self.cell_shape == 'hexahedral'
        self.cache_all_geometry_arrays()
        assert self.faces_per_cell_cl is not None and self.nodes_per_face_cl is not None
        assert self.faces_per_cell_cl[0] == 6 and np.all(self.faces_per_cell_cl[1:] - self.faces_per_cell_cl[:-1] == 6)
        assert self.nodes_per_face_cl[0] == 4 and np.all(self.nodes_per_face_cl[1:] - self.nodes_per_face_cl[:-1] == 4)

    def corner_points(self, cell):
        """Returns corner points (nodes) of a single cell.

        arguments:
           cell (int): the index of the cell for which the corner points are required

        returns:
           numpy float array of shape (8, 3) being the xyz points of 8 nodes defining a single hexahedral cell

        note:
           if this hexa grid has been created using the from_unsplit_grid class method, then the result can be
           reshaped to (2, 2, 2, 3) for corner points compatible with those used by the Grid class
        """

        self.cache_all_geometry_arrays()
        return self.points_cached[self.distinct_node_indices_for_cell(cell)]

    def volume(self, cell, required_uom = None):
        """Returns the volume of a single cell.

        arguments:
           cell (int): the index of the cell for which the volume is required

        returns:
           float being the volume of the hexahedral cell

        note:
            if required_uom is not specified, returned units will be cube of crs units if xy & z are the same
            and either 'm' or 'ft', otherwise 'm3' will be used
        """

        self._set_crs_handedness()
        apex = self.cell_centre_point(cell)
        v = 0.0
        faces, handednesses = self.face_indices_and_handedness_for_cell(cell)

        for face_index, handedness in zip(faces, handednesses):
            nodes = self.node_indices_for_face(face_index)
            abcd = self.points_cached[nodes]
            assert abcd.shape == (4, 3)
            v += vol.pyramid_volume(apex,
                                    abcd[0],
                                    abcd[1],
                                    abcd[2],
                                    abcd[3],
                                    crs_is_right_handed = (self.crs_is_right_handed == handedness))

        return self.adjusted_volume(v, required_uom = required_uom)

    # todo: add hexahedral specific method for centre_point()?
    # todo: also add other methods equivalent to those in Grid class



"""Function for populating an empty blocked well from its trajectory and a grid."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.crs as rqc
import resqpy.grid as grr
import resqpy.surface as rqs
import resqpy.well as rqw
import resqpy.grid_surface as rqgs
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec


def populate_blocked_well_from_trajectory(blocked_well,
                                          grid,
                                          active_only = False,
                                          quad_triangles = True,
                                          lazy = False,
                                          use_single_layer_tactics = True,
                                          check_for_reentry = True):
    """Populate an empty blocked well object based on the intersection of its trajectory with a grid.

    arguments:
       blocked_well (resqpy.well.BlockedWell object): a largely empty blocked well object to be populated by this
          function; note that the trajectory attribute must be set before calling this function
       grid (resqpy.grid.Grid object): the grid to intersect the well trajectory with
       active_only (boolean, default False): if True, intervals which cover inactive cells will be set as
          unblocked intervals; if False, intervals covering any cell will be set as blocked intervals
       quad_triangles (boolean, default True): if True, each cell face is represented by 4 triangles when
          calculating intersections; if False, only 2 triangles are used
       lazy (boolean, default False): if True, initial penetration must be through a top K face and blocking
          will cease as soon as the trajectory first leaves the gridded volume; if False, initial entry may be
          through any external face or a fault face and re-entry will be handled
       use_single_layer_tactics (boolean, default True): if True and not lazy and the grid does not have k gaps,
          fault planes and grid sidewall are initially treated as if the grid were a single layer, when looking
          for penetrations from outwith the grid
       check_for_reentry (boolean, default True): if True, the trajectory is tracked after leaving the grid through
          the outer skin (eg. base reservoir) in case of re-entry; if False, blocking stops upon the first exit
          of the trajectory through the skin; ignored (treated as False) if lazy is True

    returns:
       the blocked well object (same object as passed in) if successful; None if unsuccessful

    notes:
       the blocked_well trajectory attribute must be set before calling this function;
       grids with k gaps might result in very slow processing;
       the function represents a cell face as 2 or 4 triangles rather than a bilinear patch; setting quad_triangles
       False is not recommended as the 2 triangle formulation gives a non-unique representation of a face (though
       the code is designed to use the same representation for a shared face between neighbouring cells);
       where non-planar faults exist, the triangulation of faces may result in a small misalignment of abutted faces
       between the opposing sides of the fault; this could potentially result in an extra, small, unblocked interval
       being introduced as an artefact between the exit point of one cell and the entry point into the abutted cell
    """

    assert isinstance(blocked_well, rqw.BlockedWell)
    assert isinstance(blocked_well.trajectory, rqw.Trajectory)
    assert grid is not None

    flavour = grr.grid_flavour(grid.root)
    if not flavour.startswith('Ijk'):
        raise NotImplementedError('well blocking only implemented for IjkGridRepresentation')
    is_regular = (flavour == 'IjkBlockGrid') and hasattr(grid, 'is_aligned') and grid.is_aligned
    log.debug(f"well blocking: grid {'is' if is_regular else 'is not'} regular and aligned")

    if grid.k_gaps:
        use_single_layer_tactics = False
        log.debug('skin single layer tactics disabled')

    grid_crs = rqc.Crs(grid.model, uuid = grid.crs_uuid)
    trajectory = __trajectory_init(blocked_well, grid, grid_crs)
    traj_xyz = trajectory.control_points

    if not rqgs.trajectory_grid_overlap(trajectory, grid):
        log.error(f'no overlap of trajectory xyz box with grid for trajectory uuid: {trajectory.uuid}')
        return None
    grid_box = grid.xyz_box(lazy = False, local = True).copy()
    if grid_crs.z_inc_down:
        z_sign = 1.0
        grid_top_z = grid_box[0, 2]
    else:
        z_sign = -1.0
        grid_top_z = -grid_box[1, 2]  # maximum is least negative, ie. shallowest

    assert z_sign * traj_xyz[0, 2] < grid_top_z, 'trajectory does not start above top of grid'  # min z
    if z_sign * traj_xyz[-1, 2] < grid_top_z:
        log.warning('end of trajectory (TD) is above top of grid')

    # scan down wellbore till top of grid reached
    knot = 0
    knot_count = trajectory.knot_count - 1
    while knot < knot_count and z_sign * traj_xyz[knot + 1, 2] < grid_top_z:
        knot += 1
    log.debug(f'skipped trajectory to knot: {knot}')
    if knot == knot_count:
        log.warning('no well blocking due to entire trajectory being shallower than top of grid')
        return None  # entire well is above grid

    if lazy:
        # search for intersection with top of grid: will fail if well comes in from the side (or from below!) or
        # penetrates through fault
        xyz, entry_knot, col_ji0 = rqgs.find_first_intersection_of_trajectory_with_layer_interface(
            trajectory,
            grid,
            k0 = 0,
            ref_k_faces = 'top',
            start = knot,
            heal_faults = False,
            quad_triangles = quad_triangles,
            is_regular = is_regular)
        log.debug(f'lazy top intersection x,y,z: {xyz}; knot: {entry_knot}; col j0,i0: {col_ji0[0]}, {col_ji0[1]}')
        if xyz is None:
            log.error('failed to lazily find intersection of trajectory with top surface of grid')
            return None
        cell_kji0 = np.array((0, col_ji0[0], col_ji0[1]), dtype = int)
        axis = 0
        polarity = 0

    else:  # not lazy
        # note: xyz and entry_fraction might be slightly off when penetrating a skewed fault plane – deemed immaterial
        # for real cases
        skin = grid.skin(use_single_layer_tactics = use_single_layer_tactics, is_regular = is_regular)
        xyz, cell_kji0, axis, polarity, entry_knot = skin.find_first_intersection_of_trajectory(trajectory)
        if xyz is None:
            log.error('failed to find intersection of trajectory with outer skin of grid')
            return None
        else:
            log.debug(f"skin intersection x,y,z: {xyz}; knot: {entry_knot}; cell kji0: {cell_kji0}; face: "
                      f"{'KJI'[axis]}{'-+'[polarity]}")
            cell_kji0 = np.array(cell_kji0, dtype = int)

    previous_kji0 = cell_kji0.copy()
    previous_kji0[axis] += polarity * 2 - 1  # note: previous may legitimately be 'beyond' edge of grid
    entry_fraction = __segment_fraction(traj_xyz, entry_knot, xyz)
    log.debug(f'initial previous kji0: {previous_kji0}')
    next_cell_info = __find_next_cell(grid,
                                      previous_kji0,
                                      axis,
                                      1 - polarity,
                                      trajectory,
                                      entry_knot,
                                      entry_fraction,
                                      xyz,
                                      check_for_reentry,
                                      treat_skin_as_fault = use_single_layer_tactics,
                                      lazy = lazy,
                                      use_single_layer_tactics = use_single_layer_tactics)
    log.debug(f'initial next cell info: {next_cell_info}')
    node_mds_list = [__back_calculated_md(trajectory, entry_knot, entry_fraction)]
    node_count = 1
    grid_indices_list = []
    cell_count = 0
    cell_indices_list = []
    face_pairs_list = []
    kissed = np.zeros(grid.extent_kji, dtype = bool)
    sample_test = np.zeros(grid.extent_kji, dtype = bool)

    while next_cell_info is not None:
        entry_shared, kji0, entry_axis, entry_polarity, entry_knot, entry_fraction, entry_xyz = next_cell_info

        # log.debug('next cell entry x,y,z: ' + str(entry_xyz) + '; knot: ' + str(entry_knot) + '; cell kji0: ' +
        #           str(kji0) + '; face: ' + 'KJI'[entry_axis] + '-+'[entry_polarity])

        if not entry_shared:
            log.debug('adding unblocked interval')
            node_mds_list.append(__back_calculated_md(trajectory, entry_knot, entry_fraction))
            node_count += 1
            grid_indices_list.append(-1)
            kissed[:] = False
            sample_test[:] = False

        exit_xyz, exit_knot, exit_axis, exit_polarity = rqgs.find_first_intersection_of_trajectory_with_cell_surface(
            trajectory, grid, kji0, entry_knot, start_xyz = entry_xyz, nudge = 0.01, quad_triangles = True)

        #  if exit_xyz is None:
        #      log.debug('no exit')
        # else:
        #     log.debug('cell exit x,y,z: ' + str(exit_xyz) + '; knot: ' + str(exit_knot) + '; face: ' +
        #               'KJI'[exit_axis] + '-+'[exit_polarity])

        if exit_xyz is None:
            if rqgs.point_is_within_cell(traj_xyz[-1], grid, kji0):
                # well terminates within cell: add termination blocked interval (or unblocked if inactive cell)
                log.debug(f'adding termination interval for cell {kji0}')
                # todo: check for inactive cell and make an unblocked interval instead, if required
                node_mds_list.append(trajectory.measured_depths[-1])
                node_count += 1
                grid_indices_list.append(0)
                cell_indices_list.append(grid.natural_cell_index(kji0))
                cell_count += 1
                face_pairs_list.append(((entry_axis, entry_polarity), (-1, -1)))
                break
            else:
                next_cell_info = __forward_nudge(axis, entry_axis, entry_knot, entry_polarity, entry_xyz, grid, kissed,
                                                 kji0, previous_kji0, sample_test, traj_xyz, trajectory,
                                                 use_single_layer_tactics)
            if next_cell_info is None:
                log.warning('well blocking got stuck – cells probably omitted at tail of well')
        #      if exit_xyz is not None:  # usual well-behaved case or non-standard due to kiss
        else:
            exit_fraction = __segment_fraction(traj_xyz, exit_knot, exit_xyz)
            log.debug(f'adding blocked interval for cell kji0: {kji0}')
            # todo: check for inactive cell and make an unblocked interval instead, if required
            node_mds_list.append(__back_calculated_md(trajectory, exit_knot, exit_fraction))
            node_count += 1
            grid_indices_list.append(0)
            cell_indices_list.append(grid.natural_cell_index(kji0))
            cell_count += 1
            face_pairs_list.append(((entry_axis, entry_polarity), (exit_axis, exit_polarity)))

            previous_kji0 = kji0
            # log.debug(f'previous kji0 set to {previous_kji0}')
            next_cell_info = __find_next_cell(grid,
                                              kji0,
                                              exit_axis,
                                              exit_polarity,
                                              trajectory,
                                              exit_knot,
                                              exit_fraction,
                                              exit_xyz,
                                              check_for_reentry,
                                              treat_skin_as_fault = use_single_layer_tactics,
                                              lazy = lazy,
                                              use_single_layer_tactics = use_single_layer_tactics)
            kissed[:] = False
            sample_test[:] = False

    if node_count < 2:
        log.warning('no nodes found during attempt to block well')
        return None

    assert node_count > 1
    assert len(node_mds_list) == node_count
    assert len(grid_indices_list) == node_count - 1
    assert cell_count < node_count
    assert len(cell_indices_list) == cell_count
    assert len(face_pairs_list) == cell_count

    blocked_well.node_mds = np.array(node_mds_list, dtype = float)
    blocked_well.node_count = node_count
    blocked_well.grid_indices = np.array(grid_indices_list, dtype = int)
    blocked_well.cell_indices = np.array(cell_indices_list, dtype = int)
    blocked_well.face_pair_indices = np.array(face_pairs_list, dtype = int)
    blocked_well.cell_count = cell_count
    blocked_well.grid_list = [grid]

    assert cell_count == (node_count - np.count_nonzero(blocked_well.grid_indices == -1) - 1)

    log.info(f'{cell_count} cell{__pl(cell_count)} blocked for well trajectory uuid: {trajectory.uuid}')

    return blocked_well


def generate_surface_for_blocked_well_cells(blocked_well,
                                            combined = True,
                                            active_only = False,
                                            min_k0 = 0,
                                            max_k0 = None,
                                            depth_limit = None,
                                            quad_triangles = True):
    """Returns a surface or list of surfaces representing the faces of the cells visited by the well."""

    assert blocked_well is not None and type(blocked_well) is rqw.BlockedWell
    cells_kji0, grid_list = blocked_well.cell_indices_and_grid_list()
    assert len(cells_kji0) == len(grid_list)
    cell_count = len(cells_kji0)
    if cell_count == 0:
        return None

    surface_list = []
    if combined:
        composite_cp = np.zeros((cell_count, 2, 2, 2, 3))

    cell_p = 0
    for cell in range(cell_count):
        grid = grid_list[cell]
        cell_kji0 = cells_kji0[cell]
        if active_only and grid.inactive is not None and grid.inactive[cell_kji0]:
            continue
        if cell_kji0[0] < min_k0 or (max_k0 is not None and cell_kji0[0] > max_k0):
            continue
        if depth_limit is not None and grid.centre_point(cell_kji0) > depth_limit:
            continue
        cp = grid.corner_points(cell_kji0)
        if combined:
            composite_cp[cell_p] = cp
            cell_p += 1
        else:
            cs = rqs.Surface(blocked_well.model)
            cs.set_to_single_cell_faces_from_corner_points(cp, quad_triangles = quad_triangles)
            surface_list.append(cs)

    if cell_p == 0 and len(surface_list) == 0:
        return None

    if combined:
        cs = rqs.Surface(blocked_well.model)
        cs.set_to_multi_cell_faces_from_corner_points(composite_cp[:cell_p])
        return cs

    return surface_list


def __find_next_cell(
    grid,
    previous_kji0,
    axis,
    polarity,
    trajectory,
    segment,
    seg_fraction,
    xyz,
    check_for_reentry,
    treat_skin_as_fault = False,
    lazy = False,
    use_single_layer_tactics = True,
):
    # returns for next cell entry: (shared transit point bool, kji0, axis, polarity, segment, seg_fraction, xyz)
    # or single None if no next cell identified (end of trajectory beyond edge of grid)
    # take care of: edges of model; pinchouts; faults, (k gaps), exact edge or corner crossings
    # note: identified cell may be active or inactive: calling code to handle that
    # note: for convenience, previous_kji0 may lie just outside the extent of the grid
    # note: polarity is relative to previous cell, so its complement applies to the next cell
    #      log.debug('finding next cell with previous kji0: ' + str(previous_kji0) + '; exit: ' + 'kji'[axis] + '-+'[polarity])
    kji0 = np.array(previous_kji0, dtype = int)
    polarity_sign = 2 * polarity - 1
    kji0[axis] += polarity_sign
    # if gone beyond external skin of model, return None
    if np.any(kji0 < 0) or np.any(kji0 >= grid.extent_kji) or (grid.k_gaps and axis == 0 and (
        (polarity == 1 and previous_kji0[0] >= 0 and grid.k_gap_after_array[previous_kji0[0]]) or
        (polarity == 0 and kji0[0] < grid.nk - 1 and grid.k_gap_after_array[kji0[0]]))):
        if check_for_reentry and not lazy:
            skin = grid.skin(use_single_layer_tactics = use_single_layer_tactics)
            # nudge in following will be problematic if gap has zero (or tiny) thickness at this location
            xyz_r, cell_kji0, axis, polarity, segment = skin.find_first_intersection_of_trajectory(trajectory,
                                                                                                   start = segment,
                                                                                                   start_xyz = xyz,
                                                                                                   nudge = +0.05)
            if xyz_r is None:
                log.debug('no skin re-entry found after exit through skin')
                return None
            log.debug(f"skin re-entry after skin exit: kji0: {cell_kji0}; face: {'KJI'[axis]}{'-+'[polarity]}")
            seg_fraction = __segment_fraction(trajectory.control_points, segment, xyz_r)
            return False, cell_kji0, axis, polarity, segment, seg_fraction, xyz_r
        else:
            return None
    # pre-assess split pillar (fault) situation
    faulted = False
    if axis != 0:
        if grid.has_split_coordinate_lines:
            faulted = grid.is_split_column_face(kji0[1], kji0[2], axis, 1 - polarity)
        if not faulted and treat_skin_as_fault:
            faulted = (kji0[axis] == 0 and polarity == 1) or (kji0[axis] == grid.extent_kji[axis] - 1 and polarity == 0)
    # handle the simplest case of a well behaved k neighbour or unsplit j or i neighbour
    if not grid.pinched_out(cell_kji0 = kji0, cache_pinchout_array = True) and not faulted:
        return True, kji0, axis, 1 - polarity, segment, seg_fraction, xyz
    if faulted:
        return __faulted(trajectory, grid, segment, kji0, axis, polarity, xyz, lazy, use_single_layer_tactics,
                         previous_kji0)
    else:
        # skip pinched out cells
        pinchout_skip_sign = -1 if axis == 0 and polarity == 0 else 1
        while grid.pinched_out(cell_kji0 = kji0, cache_pinchout_array = True):
            kji0[0] += pinchout_skip_sign
            if not (np.all(kji0 >= 0) and np.all(kji0 < grid.extent_kji)) or (grid.k_gaps and axis == 0 and (
                (polarity == 0 and grid.k_gap_after_array[kji0[0]]) or
                (polarity == 1 and grid.k_gap_after_array[kji0[0] - 1]))):
                log.debug(
                    f"trajectory reached edge of model {'or k gap ' if grid.k_gaps and axis == 0 else ''}at exit from cell kji0: {previous_kji0}"
                )
                if lazy:
                    return None
                skin = grid.skin(use_single_layer_tactics = use_single_layer_tactics)
                # nudge in following will be problematic if gap has zero (or tiny) thickness at this location
                xyz_r, cell_kji0, axis, polarity, segment = skin.find_first_intersection_of_trajectory(trajectory,
                                                                                                       start = segment,
                                                                                                       start_xyz = xyz,
                                                                                                       nudge = +0.01)
                if xyz_r is None:
                    return None  # no re-entry found after exit through skin
                seg_fraction = __segment_fraction(trajectory.control_points, segment, xyz_r)
                return False, cell_kji0, axis, polarity, segment, seg_fraction, xyz_r
        return True, kji0, axis, 1 - polarity, segment, seg_fraction, xyz


def __faulted(trajectory, grid, segment, kji0, axis, polarity, xyz, lazy, use_single_layer_tactics, previous_kji0):
    # look for intersections with column face
    xyz_f, k0 = rqgs.find_intersection_of_trajectory_interval_with_column_face(trajectory,
                                                                               grid,
                                                                               segment,
                                                                               kji0[1:],
                                                                               axis,
                                                                               1 - polarity,
                                                                               start_xyz = xyz,
                                                                               nudge = -0.1,
                                                                               quad_triangles = True)
    if xyz_f is not None and k0 is not None:
        kji0[0] = k0
        seg_fraction = __segment_fraction(trajectory.control_points, segment, xyz_f)
        return vec.isclose(xyz, xyz_f, tolerance = 0.001), kji0, axis, 1 - polarity, segment, seg_fraction, xyz_f
    log.debug('failed to find entry point in column face after crossing fault; checking entire cross section')
    x_sect_surf = rqgs.generate_torn_surface_for_x_section(grid,
                                                           'KJI'[axis],
                                                           ref_slice0 = kji0[axis],
                                                           plus_face = (polarity == 0),
                                                           quad_triangles = True,
                                                           as_single_layer = False)
    xyz_f, segment_f, tri_index_f = rqgs.find_first_intersection_of_trajectory_with_surface(trajectory,
                                                                                            x_sect_surf,
                                                                                            start = segment,
                                                                                            start_xyz = xyz,
                                                                                            nudge = -0.1)
    if xyz_f is not None:
        # back out cell info from triangle index; note 'column_from...' is actually x_section cell face
        k0, j_or_i0 = x_sect_surf.column_from_triangle_index(tri_index_f)
        kji0[0] = k0
        kji0[3 - axis] = j_or_i0
        seg_fraction = __segment_fraction(trajectory.control_points, segment_f, xyz_f)
        return vec.isclose(xyz, xyz_f, tolerance = 0.001), kji0, axis, 1 - polarity, segment_f, seg_fraction, xyz_f
    log.debug("failed to find entry point in cross section after crossing fault"
              f"{'' if lazy else '; checking for skin re-entry'}")
    if lazy:
        return None
    skin = grid.skin(use_single_layer_tactics = use_single_layer_tactics)
    # following is problematic due to skewed fault planes
    xyz_r, cell_kji0, axis, polarity, segment = skin.find_first_intersection_of_trajectory(trajectory,
                                                                                           start = segment,
                                                                                           start_xyz = xyz,
                                                                                           nudge = -0.1,
                                                                                           exclude_kji0 = previous_kji0)
    if xyz_r is None:
        log.warning('no skin re-entry found after exit through fault face')
        return None
    log.debug(f"skin re-entry after fault face exit: kji0: {cell_kji0}; face: {'KJI'[axis]}{'-+'[polarity]}")
    seg_fraction = __segment_fraction(trajectory.control_points, segment, xyz_r)
    return False, cell_kji0, axis, polarity, segment, seg_fraction, xyz_r


def __back_calculated_md(trajectory, segment, fraction):
    base_md = trajectory.measured_depths[segment]
    return base_md + fraction * (trajectory.measured_depths[segment + 1] - base_md)


def __forward_nudge(axis, entry_axis, entry_knot, entry_polarity, entry_xyz, grid, kissed, kji0, previous_kji0,
                    sample_test, traj_xyz, trajectory, use_single_layer_tactics):
    next_cell_info = None
    # trajectory has kissed corner or edge of cell and nudge has gone outside cell
    # nudge forward and look for point inclusion in possible neighbours
    log.debug(f"kiss detected at cell kji0 {kji0} {'KJI'[entry_axis]}{'-+'[entry_polarity]}")
    kissed[tuple(kji0)] = True
    if np.all(previous_kji0 >= 0) and np.all(previous_kji0 < grid.extent_kji):
        kissed[tuple(previous_kji0)] = True  # stops immediate revisit for kiss and tell tactics
    # setup kji offsets of neighbouring cells likely to contain sample point (could check for split
    # column faces)
    axis_a = (entry_axis + 1) % 3
    axis_b = (entry_axis + 2) % 3
    offsets_kji = np.zeros((18, 3), dtype = int)
    offsets_kji[0, axis_a] = -1
    offsets_kji[1, axis_a] = 1
    offsets_kji[2, axis_b] = -1
    offsets_kji[3, axis_b] = 1
    offsets_kji[4, axis_a] = offsets_kji[4, axis_b] = -1
    offsets_kji[5, axis_a] = -1
    offsets_kji[5, axis_b] = 1
    offsets_kji[6, axis_a] = 1
    offsets_kji[6, axis_b] = -1
    offsets_kji[7, axis_a] = offsets_kji[7, axis_b] = 1
    offsets_kji[8:16] = offsets_kji[:8]
    offsets_kji[8:16, axis] = previous_kji0[axis] - kji0[axis]
    pinchout_skip_sign = -1 if entry_axis == 0 and entry_polarity == 1 else 1
    offsets_kji[16, axis] = pinchout_skip_sign
    log.debug(f'kiss pinchout skip sign: {pinchout_skip_sign}')
    next_cell_info = __try_cell_entry(entry_knot, entry_xyz, grid, kissed, kji0, next_cell_info, offsets_kji,
                                      pinchout_skip_sign, traj_xyz, trajectory)
    if next_cell_info is not None:
        return next_cell_info
    else:
        log.debug('kiss and tell failed to find next cell entry; switching to point in cell tactics')
        # take a sample point a little ahead on the trajectory
        remaining = traj_xyz[entry_knot + 1] - entry_xyz
        remaining_length = vec.naive_length(remaining)
        if remaining_length < 0.025:
            log.debug(f'not much remaining of segnemt: {remaining_length}')
        nudge = 0.025
        if remaining_length <= nudge:
            sample_point = entry_xyz + 0.9 * remaining
        else:
            sample_point = entry_xyz + nudge * vec.unit_vector(remaining)
        sample_test[:] = False
        log.debug(f'sample point is {sample_point}')
        for try_index in range(len(offsets_kji)):
            try_kji0 = np.array(kji0, dtype = int) + offsets_kji[try_index]
            while np.all(try_kji0 >= 0) and np.all(try_kji0 < grid.extent_kji) and grid.pinched_out(
                    cell_kji0 = try_kji0):
                try_kji0[0] += pinchout_skip_sign
            if np.any(try_kji0 < 0) or np.any(try_kji0 >= grid.extent_kji):
                continue
            #                 log.debug(f'tentatively trying: {try_kji0}')
            sample_test[tuple(try_kji0)] = True
            if rqgs.point_is_within_cell(sample_point, grid, try_kji0):
                log.debug(f'sample point is in cell {try_kji0}')
                try_entry_xyz, try_entry_knot, try_entry_axis, try_entry_polarity = \
                    rqgs.find_first_intersection_of_trajectory_with_cell_surface(trajectory, grid, try_kji0, entry_knot,
                                                                            start_xyz=entry_xyz, nudge=-0.01,
                                                                            quad_triangles=True)
                if try_entry_xyz is None or try_entry_knot is None:
                    log.warning(f'failed to find entry to favoured cell {try_kji0}')
                else:
                    if np.all(try_kji0 == kji0):
                        log.debug(f'staying in cell {kji0} after kiss')
                        # TODO: sort things out to discard kiss completely
                    try_entry_fraction = __segment_fraction(traj_xyz, try_entry_knot, try_entry_xyz)
                    return vec.isclose(try_entry_xyz, entry_xyz, tolerance=0.01), try_kji0, try_entry_axis, \
                           try_entry_polarity, try_entry_knot, try_entry_fraction, try_entry_xyz
                break
        log.debug('sample point not found in immediate neighbours, switching to full column search')
        k_offset = 0
        while k_offset <= max(kji0[0], grid.nk - kji0[0] - 1):
            for k_plus_minus in [1, -1]:
                for j_offset in [0, -1, 1]:
                    for i_offset in [0, -1, 1]:
                        try_kji0 = kji0 + np.array((k_offset * k_plus_minus, j_offset, i_offset))
                        if np.any(try_kji0 < 0) or np.any(try_kji0 >= grid.extent_kji) or sample_test[tuple(try_kji0)]:
                            continue
                        sample_test[tuple(try_kji0)] = True
                        if rqgs.point_is_within_cell(sample_point, grid, try_kji0):
                            log.debug(f'sample point is in cell {try_kji0}')
                            try_entry_xyz, try_entry_knot, try_entry_axis, try_entry_polarity = \
                                rqgs.find_first_intersection_of_trajectory_with_cell_surface(
                                    trajectory, grid, try_kji0, entry_knot, start_xyz=entry_xyz,
                                    nudge=-0.01, quad_triangles=True)
                            if try_entry_xyz is None or try_entry_knot is None:
                                log.warning(f'failed to find entry to column sampled cell {try_kji0}')
                                return __no_success(entry_knot, entry_xyz, grid, next_cell_info, previous_kji0,
                                                    traj_xyz, trajectory, use_single_layer_tactics)
                            else:
                                try_entry_fraction = __segment_fraction(traj_xyz, try_entry_knot, try_entry_xyz)
                                return (vec.isclose(try_entry_xyz, entry_xyz,
                                                    tolerance = 0.01), try_kji0, try_entry_axis, try_entry_polarity,
                                        try_entry_knot, try_entry_fraction, try_entry_xyz)
            k_offset += 1
        return __no_success(entry_knot, entry_xyz, grid, next_cell_info, previous_kji0, traj_xyz, trajectory,
                            use_single_layer_tactics)


def __no_success(entry_knot, entry_xyz, grid, next_cell_info, previous_kji0, traj_xyz, trajectory,
                 use_single_layer_tactics):
    log.debug('no success during full column search')
    if np.any(previous_kji0 == 0) or np.any(previous_kji0 == grid.extent_kji - 1):
        log.debug('looking for skin re-entry after possible kissing exit')
        skin = grid.skin(use_single_layer_tactics = use_single_layer_tactics)
        try_entry_xyz, try_kji0, try_entry_axis, try_entry_polarity, try_entry_knot = \
            skin.find_first_intersection_of_trajectory(trajectory, start=entry_knot,
                                                       start_xyz=entry_xyz, nudge=+0.1)
        if try_entry_xyz is not None:
            log.debug('skin re-entry found')
            try_entry_fraction = __segment_fraction(traj_xyz, try_entry_knot, try_entry_xyz)
            next_cell_info = vec.isclose(try_entry_xyz, entry_xyz, tolerance=0.01), try_kji0, try_entry_axis, \
                             try_entry_polarity, try_entry_knot, try_entry_fraction, try_entry_xyz
    return next_cell_info


def __try_cell_entry(entry_knot, entry_xyz, grid, kissed, kji0, next_cell_info, offsets_kji, pinchout_skip_sign,
                     traj_xyz, trajectory):
    for try_index in range(len(offsets_kji)):
        try_kji0 = kji0 + offsets_kji[try_index]
        while np.all(try_kji0 >= 0) and np.all(try_kji0 < grid.extent_kji) and grid.pinched_out(cell_kji0 = try_kji0):
            try_kji0[0] += pinchout_skip_sign
        if np.any(try_kji0 < 0) or np.any(try_kji0 >= grid.extent_kji) or kissed[tuple(try_kji0)]:
            continue
        # use tiny negative nudge to look for entry into try cell with current segment; check that entry
        # point is close
        try_entry_xyz, try_entry_knot, try_entry_axis, try_entry_polarity = \
            rqgs.find_first_intersection_of_trajectory_with_cell_surface(trajectory,
                                                                    grid,
                                                                    try_kji0,
                                                                    entry_knot,
                                                                    start_xyz=entry_xyz,
                                                                    nudge=-0.01,
                                                                    quad_triangles=True)
        if try_entry_xyz is not None and vec.isclose(try_entry_xyz, entry_xyz, tolerance = 0.02):
            log.debug(f'try accepted for cell: {try_kji0}')
            try_entry_fraction = __segment_fraction(traj_xyz, try_entry_knot, try_entry_xyz)
            # replace the next cell entry info
            next_cell_info = True, try_kji0, try_entry_axis, try_entry_polarity, try_entry_knot, try_entry_fraction, \
                             try_entry_xyz
            break
    return next_cell_info


def __segment_fraction(trajectory_xyz, segment, xyz):
    # returns fraction of way along segment that point xyz lies
    segment_vector = trajectory_xyz[segment + 1] - trajectory_xyz[segment]
    segment_length = vec.naive_length(segment_vector)  # note this is length in straight line, rather than diff in mds
    return vec.naive_length(xyz - trajectory_xyz[segment]) / segment_length


def __trajectory_init(blocked_well, grid, grid_crs):
    if bu.matching_uuids(blocked_well.trajectory.crs_uuid, grid.crs_uuid):
        trajectory = blocked_well.trajectory
        assert grid_crs == rqc.Crs(blocked_well.model, uuid = trajectory.crs_uuid)
    else:
        # create a local trajectory object for crs alignment of control points with grid
        # NB. temporary objects, relationships left in a mess
        # model = rq.Model(create_basics = True)
        log.debug(f'creating temp trajectory in grid crs for well {rqw.well_name(blocked_well.trajectory)}')
        model = blocked_well.trajectory.model
        trajectory = rqw.Trajectory(model, uuid = blocked_well.trajectory.uuid)
        assert trajectory is not None and trajectory.control_points is not None
        traj_crs = rqc.Crs(model, uuid = trajectory.crs_uuid)
        traj_crs.convert_array_to(grid_crs, trajectory.control_points)  # trajectory xyz converted in situ to grid's crs
        trajectory.crs_uuid = grid_crs.uuid
        log.debug(
            f'temp traj xyz box: {np.min(trajectory.control_points, axis = 0)} to {np.max(trajectory.control_points, axis = 0)}'
        )
        # note: any represented interpretation object will not be present in the temporary model
    return trajectory


def __pl(n, use_es = False):
    if n == 1:
        return ''
    return 'es' if use_es else 's'



"""Trajectory class."""

# todo: create a trajectory from a deviation survey, assuming minimum curvature

# Nexus is a registered trademark of the Halliburton Company
# RMS and ROXAR are registered trademarks of Roxar Software Solutions AS, an Emerson company

import logging

log = logging.getLogger(__name__)

import math as maths
import numpy as np
import pandas as pd
from functools import partial

import resqpy.lines as rql
import resqpy.olio.uuid as bu
import resqpy.olio.vector_utilities as vec
import resqpy.olio.wellspec_keywords as wsk
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo
import resqpy.weights_and_measures as bwam
import resqpy.well
import resqpy.well._md_datum as rqmdd
import resqpy.well._wellbore_frame as rqwbf
import resqpy.well._deviation_survey as rqds
import resqpy.well.well_utils as rqwu
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class Trajectory(BaseResqpy):
    """Class for RESQML Wellbore Trajectory Representation (Geometry).

    note:
       resqml allows trajectory to have different crs to the measured depth datum crs;
       however, this code requires the trajectory to be in the same crs as the md datum
    """

    resqml_type = 'WellboreTrajectoryRepresentation'
    well_name = rqo.alias_for_attribute("title")

    def __init__(self,
                 parent_model,
                 uuid = None,
                 crs_uuid = None,
                 md_datum = None,
                 deviation_survey = None,
                 data_frame = None,
                 grid = None,
                 cell_kji0_list = None,
                 wellspec_file = None,
                 spline_mode = 'cube',
                 ascii_trajectory_file = None,
                 survey_file_space_separated = False,
                 length_uom = None,
                 md_domain = None,
                 represented_interp = None,
                 well_name = None,
                 set_tangent_vectors = False,
                 hdf5_source_model = None,
                 originator = None,
                 extra_metadata = None):
        """Creates a new trajectory object and optionally loads it from xml, deviation survey, pandas dataframe, or

        ascii file.

        arguments:
           parent_model (model.Model object): the model which the new trajectory belongs to
           uuid (UUID, optional): if present, the Trajectory is initialised from xml for an existing RESQML object
              and the remaining arguments are mostly ignored
           crs_uuid (UUID, optional): the uuid of a Crs object to use when generating a new trajectory
           md_datum (MdDatum object): the datum that the depths for this trajectory are measured from;
              not used if uuid is not None
           deviation_survey (DeviationSurvey object, optional): if present and uuid is None
              then the trajectory is derived from the deviation survey based on minimum curvature
           data_frame (optional): a pandas dataframe with columns 'MD', 'X', 'Y' and 'Z', holding
              the measured depths, and corresponding node locations; ignored if uuid is not None
           grid (grid.Grid object, optional): only required if initialising from a list of cell indices;
              ignored otherwise
           cell_kji0_list (numpy int array of shape (N, 3)): ordered list of cell indices to be visited by
              the trajectory; ignored if uuid is not None
           wellspec_file (string, optional): name of an ascii file containing Nexus WELLSPEC data; well_name
              and length_uom arguments must be passed
           spline_mode (string, default 'cube'): one of 'none', 'linear', 'square', or 'cube'; affects spline
              tangent generation; only relevant if initialising from list of cells
           ascii_trajectory_file (string): filename of an ascii file holding the trajectory
              in a tabular form; ignored if uuid is not None
           survey_file_space_separated (boolean, default False): if True, deviation survey file is
              space separated; if False, comma separated (csv); ignored unless loading from survey file
           length_uom (string, default 'm'): a resqml length unit of measure applicable to the
              measured depths; should be 'm' or 'ft'
           md_domain (string, optional): if present, must be 'logger' or 'driller'; the source of the original
              deviation data; ignored if uuid is not None
           represented_interp (wellbore interpretation object, optional): if present, is noted as the wellbore
              interpretation object which this trajectory relates to; ignored if uuid is not None
           well_name (string, optional): used as citation title
           set_tangent_vectors (boolean, default False): if True and tangent vectors are not loaded then they will
              be computed from the control points
           hdf5_source_model (model.Model, optional): if present this model is used to determine the hdf5 file
              name from which to load the trajectory's array data; if None, the parent_model is used as usual
           originator (str, optional): the name of the person creating the trajectory, defaults to login id;
              ignored if uuid is not None
           extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the trajectory;
              ignored if uuid is not None

        returns:
           the newly created wellbore trajectory object

        notes:
           if starting from a deviation survey file, there are two routes: create a deviation survey object first,
           using the azimuth and inclination data, then generate a trajectory from that based on minimum curvature;
           or, create a trajectory directly using X, Y, Z data from the deviation survey file (ie. minimum
           curvature or other algorithm already applied externally);
           if not loading from xml, then the crs is set to that used by the measured depth datum, or if that is not
           available then the default crs for the model

        :meta common:
        """

        self.crs_uuid = crs_uuid
        self.title = well_name
        self.start_md = None
        self.finish_md = None
        self.md_uom = length_uom
        self.md_domain = md_domain
        self.md_datum = md_datum  # md datum is an object in its own right, with a related crs!
        # parametric line geometry elements
        self.knot_count = None
        self.line_kind_index = None
        # 0 for vertical
        # 1 for linear spline
        # 2 for natural cubic spline
        # 3 for cubic spline
        # 4 for z linear cubic spline
        # 5 for minimum-curvature spline   # in practice this is the value actually used  in datasets
        # (-1) for null: no line
        self.measured_depths = None  # known as control point parameters in the parametric line geometry
        self.control_points = None  # xyz array of shape (knot_count, 3)
        self.tangent_vectors = None  # optional xyz tangent vector array, if present has same shape as control points)
        self.deviation_survey = deviation_survey  # optional related deviation survey
        self.wellbore_interpretation = represented_interp
        self.wellbore_feature = None
        self.feature_and_interpretation_to_be_written = False
        # todo: parent intersection for multi-lateral wells
        # todo: witsml trajectory reference (optional)

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = well_name,
                         originator = originator,
                         extra_metadata = extra_metadata)

        if self.root is not None:
            if set_tangent_vectors and type(self.knot_count) is int and self.tangent_vectors is None:
                if self.knot_count > 1:
                    self.set_tangents()
            return

        # Using dictionary mapping to replicate a switch statement. The init_function key is chosen based on the
        # data source and the correct function is then called based on the init_function_dict
        init_function_dict = {
            'deviation_survey':
                partial(self.compute_from_deviation_survey,
                        method = 'minimum curvature',
                        set_tangent_vectors = set_tangent_vectors),
            'data_frame':
                partial(self.load_from_data_frame,
                        data_frame,
                        md_uom = length_uom,
                        md_datum = md_datum,
                        set_tangent_vectors = set_tangent_vectors),
            'cell_kji0_list':
                partial(self.load_from_cell_list, grid, cell_kji0_list, spline_mode, length_uom),
            'wellspec_file':
                partial(self.load_from_wellspec, grid, wellspec_file, well_name, spline_mode, length_uom),
            'ascii_trajectory_file':
                partial(self.load_from_ascii_file,
                        ascii_trajectory_file,
                        space_separated_instead_of_csv = survey_file_space_separated,
                        md_uom = length_uom,
                        md_datum = md_datum,
                        title = well_name,
                        set_tangent_vectors = set_tangent_vectors)
        }

        chosen_init_method = self.__choose_init_method(data_frame = data_frame,
                                                       cell_kji0_list = cell_kji0_list,
                                                       wellspec_file = wellspec_file,
                                                       deviation_survey_file = ascii_trajectory_file)

        try:
            init_function_dict[chosen_init_method]()
        except KeyError:
            log.warning('invalid combination of input arguments specified')

        # todo: create from already loaded deviation_survey node (ie. derive xyz points)

        if self.crs_uuid is None:
            if self.md_datum is not None:
                self.crs_uuid = self.md_datum.crs_uuid
            else:
                self.crs_uuid = self.model.crs_uuid

        if not self.title:
            self.title = 'well trajectory'

        if self.md_datum is None and self.control_points is not None:
            self.md_datum = rqmdd.MdDatum(self.model, crs_uuid = self.crs_uuid, location = self.control_points[0])

        if set_tangent_vectors and type(self.knot_count) is int and self.tangent_vectors is None:
            if self.knot_count > 1:
                self.set_tangents()

    def __choose_init_method(self, data_frame, cell_kji0_list, wellspec_file, deviation_survey_file):
        """Choose an init method based on data source."""

        if self.deviation_survey is not None:
            return 'deviation_survey'
        elif data_frame is not None:
            return 'data_frame'
        elif cell_kji0_list is not None:
            return 'cell_kji0_list'
        elif wellspec_file:
            return 'wellspec_file'
        elif deviation_survey_file:
            return 'ascii_trajectory_file'
        else:
            return None

    def iter_wellbore_frames(self):
        """Iterable of all WellboreFrames associated with a trajectory.

        Yields:
           frame: instance of :class:`resqpy.organize.WellboreFrame`

        :meta common:
        """
        uuids = self.model.uuids(obj_type = "WellboreFrameRepresentation", related_uuid = self.uuid)
        for uuid in uuids:
            yield rqwbf.WellboreFrame(self.model, uuid = uuid)

    def _load_from_xml(self):
        """Loads the trajectory object from an xml node (and associated hdf5 data)."""

        node = self.root
        assert node is not None
        self.start_md = float(rqet.node_text(rqet.find_tag(node, 'StartMd')).strip())
        self.finish_md = float(rqet.node_text(rqet.find_tag(node, 'FinishMd')).strip())
        self.md_uom = rqet.length_units_from_node(rqet.find_tag(node, 'MdUom'))
        self.md_domain = rqet.node_text(rqet.find_tag(node, 'MdDomain'))
        geometry_node = rqet.find_tag(node, 'Geometry')
        self.crs_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(geometry_node, ['LocalCrs', 'UUID']))
        self.knot_count = int(rqet.node_text(rqet.find_tag(geometry_node, 'KnotCount')).strip())
        self.line_kind_index = int(rqet.node_text(rqet.find_tag(geometry_node, 'LineKindIndex')).strip())
        mds_node = rqet.find_tag(geometry_node, 'ControlPointParameters')
        if mds_node is not None:  # not required for vertical or z linear cubic spline
            rqwu.load_hdf5_array(self, mds_node, 'measured_depths')
        control_points_node = rqet.find_tag(geometry_node, 'ControlPoints')
        rqwu.load_hdf5_array(self, control_points_node, 'control_points', tag = 'Coordinates')
        tangents_node = rqet.find_tag(geometry_node, 'TangentVectors')
        if tangents_node is not None:
            rqwu.load_hdf5_array(self, tangents_node, 'tangent_vectors', tag = 'Coordinates')
        relatives_model = self.model  # if hdf5_source_model is None else hdf5_source_model
        # md_datum - separate part, referred to in this tree
        md_datum_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(node, ['MdDatum', 'UUID']))
        assert md_datum_uuid is not None, 'failed to fetch uuid of md datum for trajectory'
        md_datum_part = relatives_model.part_for_uuid(md_datum_uuid)
        assert md_datum_part, 'md datum part not found in model'
        self.md_datum = rqmdd.MdDatum(self.model, uuid = relatives_model.uuid_for_part(md_datum_part))
        ds_uuid = bu.uuid_from_string(rqet.find_nested_tags_text(node, ['DeviationSurvey', 'UUID']))
        if ds_uuid is not None:  # this will probably not work when relatives model is different from self.model
            ds_uuid = self.model.uuid(obj_type = 'DeviationSurveyRepresentation',
                                      uuid = ds_uuid)  # check part is present
            if ds_uuid is not None:
                self.deviation_survey = rqds.DeviationSurvey(self.model, uuid = ds_uuid)
        interp_uuid = rqet.find_nested_tags_text(node, ['RepresentedInterpretation', 'UUID'])
        if interp_uuid is None:
            self.wellbore_interpretation = None
        else:
            self.wellbore_interpretation = rqo.WellboreInterpretation(self.model, uuid = interp_uuid)

    def compute_from_deviation_survey(self,
                                      survey = None,
                                      method = 'minimum curvature',
                                      md_domain = None,
                                      set_tangent_vectors = True):
        """Derive wellbore trajectory from deviation survey azimuth and inclination data."""

        if survey is None:
            assert self.deviation_survey is not None
            survey = self.deviation_survey
        else:
            self.deviation_survey = survey

        assert method in ['minimum curvature']  # if adding other methods, set line_kind_index appropriately

        self.knot_count = survey.station_count
        assert self.knot_count >= 2  # vertical well could be hamdled by allowing a single station in survey?
        self.line_kind_index = 5  # minimum curvature spline
        self.measured_depths = survey.measured_depths.copy()
        self.md_uom = survey.md_uom
        if not self.title:
            self.title = rqet.find_nested_tags_text(survey.root, ['Citation', 'Title'])
        self.start_md = self.measured_depths[0]
        self.finish_md = self.measured_depths[-1]
        if md_domain is not None:
            self.md_domain = md_domain
        self.control_points = np.empty((self.knot_count, 3))
        self.control_points[0, :] = survey.first_station
        self.__calculate_trajectory_from_inclination_and_azimuth(survey = survey)
        self.tangent_vectors = None
        if set_tangent_vectors:
            self.set_tangents()
        self.md_datum = survey.md_datum

    def __calculate_trajectory_from_inclination_and_azimuth(self, survey):
        """Calculate well trajectory from inclination and azimuth data."""

        for sp in range(1, self.knot_count):
            i1 = survey.inclinations[sp - 1]
            i2 = survey.inclinations[sp]
            az1 = survey.azimuths[sp - 1]
            az2 = survey.azimuths[sp]
            delta_md = survey.measured_depths[sp] - survey.measured_depths[sp - 1]
            assert delta_md > 0.0
            if i1 == i2 and az1 == az2:
                delta_v = delta_md * vec.unit_vector_from_azimuth_and_inclination(az1, i1)
            else:
                i1 = maths.radians(i1)
                i2 = maths.radians(i2)
                az1 = maths.radians(az1)
                az2 = maths.radians(az2)
                sin_i1 = maths.sin(i1)
                sin_i2 = maths.sin(i2)
                cos_theta = min(max(maths.cos(i2 - i1) - sin_i1 * sin_i2 * (1.0 - maths.cos(az2 - az1)), -1.0), 1.0)
                theta = maths.acos(cos_theta)
                #           theta = maths.acos(sin_i1 * sin_i2 * maths.cos(az2 - az1)  +  (maths.cos(i1) * maths.cos(i2)))
                assert theta != 0.0  # shouldn't happen as covered by if clause above
                half_rf = maths.tan(0.5 * theta) / theta
                delta_y = delta_md * half_rf * ((sin_i1 * maths.cos(az1)) + (sin_i2 * maths.cos(az2)))
                delta_x = delta_md * half_rf * ((sin_i1 * maths.sin(az1)) + (sin_i2 * maths.sin(az2)))
                delta_z = delta_md * half_rf * (maths.cos(i1) + maths.cos(i2))
                delta_v = np.array((delta_x, delta_y, delta_z))
            self.control_points[sp] = self.control_points[sp - 1] + delta_v

    def load_from_data_frame(
            self,
            data_frame,
            md_col = 'MD',
            x_col = 'X',
            y_col = 'Y',
            z_col = 'Z',
            md_uom = 'm',
            md_domain = None,
            md_datum = None,  # MdDatum object
            title = None,
            set_tangent_vectors = True):
        """Load MD and control points (xyz) data from a pandas data frame."""

        try:
            for col in [md_col, x_col, y_col, z_col]:
                assert col in data_frame.columns
            self.knot_count = len(data_frame)
            assert self.knot_count >= 2  # vertical well could be hamdled by allowing a single station in survey?
            self.line_kind_index = 5  # assume minimum curvature spline
            #         self.md_uom = bwam.p_length_unit(md_uom)
            self.md_uom = bwam.rq_length_unit(md_uom)
            start = data_frame.iloc[0]
            finish = data_frame.iloc[-1]
            if title:
                self.title = title
            self.start_md = start[md_col]
            self.finish_md = finish[md_col]
            if md_domain is not None:
                self.md_domain = md_domain
            self.measured_depths = np.empty(self.knot_count)
            self.measured_depths[:] = data_frame[md_col]
            self.control_points = np.empty((self.knot_count, 3))
            self.control_points[:, 0] = data_frame[x_col]
            self.control_points[:, 1] = data_frame[y_col]
            self.control_points[:, 2] = data_frame[z_col]
            self.tangent_vectors = None
            if set_tangent_vectors:
                self.set_tangents()
            self.md_datum = md_datum
        except Exception:
            log.exception('failed to load trajectory object from data frame')

    def load_from_cell_list(self, grid, cell_kji0_list, spline_mode = 'cube', md_uom = 'm'):
        """Loads the trajectory object based on the centre points of a list of cells."""

        assert grid is not None, 'grid argument missing for trajectory initislisation from cell list'
        cell_kji0_list = np.array(cell_kji0_list, dtype = int)
        assert cell_kji0_list.ndim == 2 and cell_kji0_list.shape[1] == 3
        assert spline_mode in ['none', 'linear', 'square', 'cube']

        cell_centres = grid.centre_point_list(cell_kji0_list)

        knot_count = len(cell_kji0_list) + 2
        self.line_kind_index = 5  # 5 means minimum curvature spline; todo: set to cubic spline value?
        self.md_uom = bwam.rq_length_unit(md_uom)
        self.start_md = 0.0
        points = np.empty((knot_count, 3))
        points[1:-1] = cell_centres
        points[0] = points[1]
        points[0, 2] = 0.0
        points[-1] = points[-2]
        points[-1, 2] *= 1.05
        if spline_mode == 'none':
            self.knot_count = knot_count
            self.control_points = points
        else:
            self.control_points = rql.spline(points, tangent_weight = spline_mode, min_subdivisions = 3)
            self.knot_count = len(self.control_points)
        self.set_measured_depths()

    def load_from_wellspec(self, grid, wellspec_file, well_name, spline_mode = 'cube', md_uom = 'm'):
        """Sets the trajectory data based on visiting the cells identified in a Nexus wellspec keyword."""

        col_list = ['IW', 'JW', 'L']
        wellspec_dict = wsk.load_wellspecs(wellspec_file, well = well_name, column_list = col_list)

        assert len(wellspec_dict) == 1, 'no wellspec data found in file ' + wellspec_file + ' for well ' + well_name

        df = wellspec_dict[well_name]
        assert len(df) > 0, 'no rows of perforation data found in wellspec for well ' + well_name

        cell_kji0_list = np.empty((len(df), 3), dtype = int)
        cell_kji0_list[:, 0] = df['L']
        cell_kji0_list[:, 1] = df['JW']
        cell_kji0_list[:, 2] = df['IW']

        self.load_from_cell_list(grid, cell_kji0_list, spline_mode, md_uom)

    def load_from_ascii_file(self,
                             trajectory_file,
                             comment_character = '#',
                             space_separated_instead_of_csv = False,
                             md_col = 'MD',
                             x_col = 'X',
                             y_col = 'Y',
                             z_col = 'Z',
                             md_uom = 'm',
                             md_domain = None,
                             md_datum = None,
                             well_col = None,
                             title = None,
                             set_tangent_vectors = True):
        """Loads the trajectory object from an ascii file with columns for MD, X, Y & Z (and optionally WELL)."""

        if not title and not self.title:
            self.title = 'well trajectory'

        try:
            df = pd.read_csv(trajectory_file,
                             comment = comment_character,
                             delim_whitespace = space_separated_instead_of_csv)
            if df is None:
                raise Exception
        except Exception:
            log.error('failed to read ascii deviation survey file ' + str(trajectory_file))
            raise

        well_col = self.__check_well_col(df = df, trajectory_file = trajectory_file, well_col = well_col)
        if title:  # filter data frame by well name
            if well_col:
                df = df[df[well_col] == title]
                if len(df) == 0:
                    log.error('no data found for well ' + str(title) + ' in file ' + str(trajectory_file))
        elif well_col is not None:
            if len(set(df[well_col])) > 1:
                raise Exception(
                    'attempt to set trajectory for unidentified well from ascii file holding data for multiple wells')
        self.load_from_data_frame(df,
                                  md_col = md_col,
                                  x_col = x_col,
                                  y_col = y_col,
                                  z_col = z_col,
                                  md_uom = md_uom,
                                  md_domain = md_domain,
                                  md_datum = md_datum,
                                  title = title,
                                  set_tangent_vectors = set_tangent_vectors)

    @staticmethod
    def __check_well_col(
        df,
        trajectory_file,
        well_col = None,
    ):
        """Verifies that a valid well_col has been supplied or can be found in the dataframe that has been generated

        from the trajectory file.
        """

        if well_col and well_col not in df.columns:
            log.warning('well column ' + str(well_col) + ' not found in ascii trajectory file ' + str(trajectory_file))
            well_col = None
        if well_col is None:
            for col in df.columns:
                if str(col).upper().startswith('WELL'):
                    well_col = col
                    break
        return well_col

    def set_tangents(self, force = False, write_hdf5 = False, weight = 'cube'):
        """Calculates tangent vectors based on control points.

        arguments:
           force (boolean, default False): if False and tangent vectors already exist then the existing ones are used;
              if True or no tangents vectors exist then they are computed
           write_hdf5 (boolean, default False): if True and new tangent vectors are computed then the array is also written
              directly to the hdf5 file
           weight (string, default 'linear'): one of 'linear', 'square', 'cube'; if linear, each tangent is the mean of the
              direction vectors of the two trjectory segments which meet at the knot; the square and cube options give
              increased weight to the direction vector of shorter segments (usually better)

        returns:
           numpy float array of shape (knot_count, 3) being the tangents in xyz, 'pointing' in the direction of increased
           knot index; the tangents are also stored as an attribute of the object

        note:
           the write_hdf5() method writes all the array data for the trajectory, including the tangent vectors; only set
           the write_hdf5 argument to this method to True if the other arrays for the trajectory already exist in the hdf5 file
        """

        if self.tangent_vectors is not None and not force:
            return self.tangent_vectors
        assert self.knot_count is not None and self.knot_count >= 2
        assert self.control_points is not None and len(self.control_points) == self.knot_count

        self.tangent_vectors = rql.tangents(self.control_points, weight = weight)

        if write_hdf5:
            h5_reg = rwh5.H5Register(self.model)
            h5_reg.register_dataset(self.uuid, 'tangentVectors', self.tangent_vectors)
            h5_reg.write(file = self.model.h5_filename(), mode = 'a')

        return self.tangent_vectors

    def dataframe(self, md_col = 'MD', x_col = 'X', y_col = 'Y', z_col = 'Z'):
        """Returns a pandas data frame containing MD and control points (xyz) data.

        note:
           set md_col to None for a dataframe containing only X, Y & Z data

        :meta common:
        """

        if md_col:
            column_list = [md_col, x_col, y_col, z_col]
        else:
            column_list = [x_col, y_col, z_col]

        data_frame = pd.DataFrame(columns = column_list)
        if md_col:
            data_frame[md_col] = self.measured_depths
        data_frame[x_col] = self.control_points[:, 0]
        data_frame[y_col] = self.control_points[:, 1]
        data_frame[z_col] = self.control_points[:, 2]
        return data_frame

    def write_to_ascii_file(self,
                            trajectory_file,
                            mode = 'w',
                            space_separated_instead_of_csv = False,
                            md_col = 'MD',
                            x_col = 'X',
                            y_col = 'Y',
                            z_col = 'Z'):
        """Writes trajectory to an ascii file.

        note:
           set md_col to None for a dataframe containing only X, Y & Z data
        """

        df = self.dataframe(md_col = md_col, x_col = x_col, y_col = y_col, z_col = z_col)
        sep = ' ' if space_separated_instead_of_csv else ','
        df.to_csv(trajectory_file, sep = sep, index = False, mode = mode)

    def xyz_for_md(self, md):
        """Returns xyz corresponding to the given measured depth; uses simple linear interpolation between knots.

        args:
           md (float): measured depth for which xyz location is required; units must be those of self.md_uom

        returns:
           triple float being x, y, z coordinates of point on trajectory corresponding to given measured depth

        note:
           the algorithm uses a simple linear interpolation between neighbouring knots (control points) on the trajectory;
           if the measured depth is less than zero or greater than the finish md, a single None is returned; if the md is
           less than the start md then a linear interpolation between the md datum location and the first knot is returned

        :meta common:
        """

        def interpolate(p1, p2, f):
            return f * p2 + (1.0 - f) * p1

        def search(md, i1, i2):
            if i2 - i1 <= 1:
                if md == self.measured_depths[i1]:
                    return self.control_points[i1]
                return interpolate(self.control_points[i1], self.control_points[i1 + 1],
                                   (md - self.measured_depths[i1]) /
                                   (self.measured_depths[i1 + 1] - self.measured_depths[i1]))
            im = i1 + (i2 - i1) // 2
            if self.measured_depths[im] >= md:
                return search(md, i1, im)
            return search(md, im, i2)

        if md < 0.0 or md > self.finish_md or md > self.measured_depths[-1]:
            return None
        if md <= self.start_md:
            if self.start_md == 0.0:
                return self.md_datum.location
            return interpolate(np.array(self.md_datum.location), self.control_points[0], md / self.start_md)
        return search(md, 0, self.knot_count - 1)

    def splined_trajectory(self,
                           well_name,
                           min_subdivisions = 1,
                           max_segment_length = None,
                           max_degrees_per_knot = 5.0,
                           use_tangents_if_present = True,
                           store_tangents_if_calculated = True):
        """Creates and returns a new Trajectory derived as a cubic spline of this trajectory.

        arguments:
           well_name (string): the name to use as the citation title for the new trajectory
           min_subdivisions (+ve integer, default 1): the minimum number of segments in the trajectory for each
              segment in this trajectory
           max_segment_length (float, optional): if present, each segment of this trajectory is subdivided so
              that the naive subdivided length is not greater than the specified length
           max_degrees_per_knot (float, default 5.0): the maximum number of degrees
           use_tangents_if_present (boolean, default False): if True, any tangent vectors in this trajectory
              are used during splining
           store_tangents_if_calculated (boolean, default True): if True any tangents calculated by the method
              are stored in the object (causing any previous tangents to be discarded); however, the new tangents
              are not written to the hdf5 file by this method

        returns:
           Trajectory object with control points lying on a cubic spline of the points of this trajectory

        notes:
           this method is typically used to smoothe an artificial or simulator trajectory;
           measured depths are re-calculated and will differ from those in this trajectory;
           unexpected behaviour may occur if the z units are different from the xy units in the crs;
           if tangent vectors for neighbouring points in this trajectory are pointing in opposite directions,
           the resulting spline is likely to be bad;
           the max_segment_length is applied when deciding how many subdivisions to make for a segment in this
           trajectory, based on the stright line segment length; segments in the resulting spline may exceed this
           length;
           similarly max_degrees_per_knot assumes a simply bend between neighbouring knots; if the position of the
           control points results in a loop, the value may be exceeded in the spline;
           the hdf5 data for the splined trajectory is not written by this method, neither is the xml created;
           no interpretation object is created by this method
           NB: direction of tangent vectors affects results, set use_tangents_if_present = False to
           ensure locally calculated tangent vectors are used
        """

        assert self.knot_count > 1 and self.control_points is not None
        assert min_subdivisions >= 1
        assert max_segment_length is None or max_segment_length > 0.0
        assert max_degrees_per_knot is None or max_degrees_per_knot > 0.0
        if not well_name:
            well_name = self.title

        tangent_vectors = self.tangent_vectors
        if tangent_vectors is None or not use_tangents_if_present:
            tangent_vectors = rql.tangents(self.control_points, weight = 'square')
            if store_tangents_if_calculated:
                self.tangent_vectors = tangent_vectors

        spline_traj = Trajectory(self.model,
                                 well_name = well_name,
                                 md_datum = self.md_datum,
                                 length_uom = self.md_uom,
                                 md_domain = self.md_domain)
        spline_traj.line_kind_index = self.line_kind_index  # not sure how we should really be setting this
        spline_traj.crs_uuid = self.crs_uuid
        spline_traj.start_md = self.start_md
        spline_traj.deviation_survey = self.deviation_survey

        spline_traj.control_points = rql.spline(self.control_points,
                                                tangent_vectors = tangent_vectors,
                                                min_subdivisions = min_subdivisions,
                                                max_segment_length = max_segment_length,
                                                max_degrees_per_knot = max_degrees_per_knot)
        spline_traj.knot_count = len(spline_traj.control_points)

        spline_traj.set_measured_depths()

        return spline_traj

    def set_measured_depths(self):
        """Sets the measured depths from the start_md value and the control points."""

        self.measured_depths = np.empty(self.knot_count)
        self.measured_depths[0] = self.start_md
        for sk in range(1, self.knot_count):
            self.measured_depths[sk] = (self.measured_depths[sk - 1] +
                                        vec.naive_length(self.control_points[sk] - self.control_points[sk - 1]))
        self.finish_md = self.measured_depths[-1]

        return self.measured_depths

    def create_feature_and_interpretation(self):
        """Instantiate new empty WellboreFeature and WellboreInterpretation objects, if a wellboreinterpretation does

        not already exist.

        Uses the trajectory citation title as the well name
        """

        log.debug("Creating a new WellboreInterpretation..")
        log.debug(f"WellboreFeature exists: {self.wellbore_feature is not None}")
        log.debug(f"WellboreInterpretation exists: {self.wellbore_interpretation is not None}")

        if self.wellbore_interpretation is None:
            log.info(f"Creating WellboreInterpretation and WellboreFeature with name {self.title}")
            self.wellbore_feature = rqo.WellboreFeature(parent_model = self.model, feature_name = self.title)
            self.wellbore_interpretation = rqo.WellboreInterpretation(parent_model = self.model,
                                                                      wellbore_feature = self.wellbore_feature)
            self.feature_and_interpretation_to_be_written = True
        else:
            raise ValueError("Cannot add WellboreFeature, trajectory already has an associated WellboreInterpretation")

    def create_xml(self,
                   ext_uuid = None,
                   wbt_uuid = None,
                   md_datum_root = None,
                   md_datum_xyz = None,
                   add_as_part = True,
                   add_relationships = True,
                   title = None,
                   originator = None):
        """Create a wellbore trajectory representation node from a Trajectory object, optionally add as part.

        notes:
          measured depth datum xml node must be in place before calling this function;
          branching well structures (multi-laterals) are supported by the resqml standard but not yet by
          this code;
          optional witsml trajectory reference not yet supported here

        :meta common:
        """

        if title:
            self.title = title
        if not self.title:
            self.title = 'wellbore trajectory'

        if ext_uuid is None:
            ext_uuid = self.model.h5_uuid()

        self.__create_wellbore_feature_and_interpretation_xml(add_as_part = add_as_part,
                                                              add_relationships = add_relationships,
                                                              originator = originator)

        if md_datum_root is None:
            md_datum_root = self.__create_md_datum_root(md_datum_xyz = md_datum_xyz)

        wbt_node = super().create_xml(originator = originator, add_as_part = False)

        self.__create_wbt_node_non_geometry_sub_elements(wbt_node = wbt_node)

        self.model.create_md_datum_reference(self.md_datum.root, root = wbt_node)

        if self.line_kind_index != 0:  # 0 means vertical well, which doesn't need a geometry

            # todo: check geometry elements for parametric curve flavours other than minimum curvature

            geom, kc_node, lki_node, cpp_node, cpp_values_node, cp_node, cp_coords_node, tv_node, tv_coords_node = self.__create_wbt_node_geometry_sub_elements(
                wbt_node = wbt_node)

            self.__get_crs_uuid()

            self.model.create_crs_reference(crs_uuid = self.crs_uuid, root = geom)

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'controlPointParameters', root = cpp_values_node)

            self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'controlPoints', root = cp_coords_node)

            if (tv_node is not None) & (tv_coords_node is not None):
                self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'tangentVectors', root = tv_coords_node)

        self.__create_deviation_survey_reference_node(wbt_node = wbt_node)

        interp_root = self.__create_wellbore_interpretation_reference_node(wbt_node = wbt_node)

        self.__add_as_part_and_add_relationships(wbt_node = wbt_node,
                                                 interp_root = interp_root,
                                                 add_as_part = add_as_part,
                                                 add_relationships = add_relationships,
                                                 ext_uuid = ext_uuid)

        return wbt_node

    def __create_wellbore_feature_and_interpretation_xml(self, add_as_part, add_relationships, originator):
        """ Create root node for WellboreFeature and WellboreInterpretation objects."""

        if self.feature_and_interpretation_to_be_written:
            if self.wellbore_interpretation is None:
                self.create_feature_and_interpretation()
            if self.wellbore_feature is not None:
                self.wellbore_feature.create_xml(add_as_part = add_as_part, originator = originator)
            self.wellbore_interpretation.create_xml(add_as_part = add_as_part,
                                                    add_relationships = add_relationships,
                                                    originator = originator)

    def __create_md_datum_root(self, md_datum_xyz):
        """ Create the root node for the MdDatum object."""

        if self.md_datum is None:
            assert md_datum_xyz is not None
            self.md_datum = rqmdd.MdDatum(self.model, location = md_datum_xyz)
        if self.md_datum.root is None:
            md_datum_root = self.md_datum.create_xml()
        else:
            md_datum_root = self.md_datum.root
        return md_datum_root

    def __create_wbt_node_non_geometry_sub_elements(self, wbt_node):
        """ Append sub-elements to the Trajectory object's root node that are unrelated to well geometry."""

        start_node = rqet.SubElement(wbt_node, ns['resqml2'] + 'StartMd')
        start_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        start_node.text = str(self.start_md)

        finish_node = rqet.SubElement(wbt_node, ns['resqml2'] + 'FinishMd')
        finish_node.set(ns['xsi'] + 'type', ns['xsd'] + 'double')
        finish_node.text = str(self.finish_md)

        md_uom = rqet.SubElement(wbt_node, ns['resqml2'] + 'MdUom')
        md_uom.set(ns['xsi'] + 'type', ns['eml'] + 'LengthUom')
        md_uom.text = bwam.rq_length_unit(self.md_uom)

        if self.md_domain:
            domain_node = rqet.SubElement(wbt_node, ns['resqml2'] + 'MdDomain')
            domain_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'MdDomain')
            domain_node.text = self.md_domain

    def __create_wbt_node_geometry_sub_elements(self, wbt_node):
        """ Append sub-elements to the Trajectory object's root node that are related to well geometry."""

        geom = rqet.SubElement(wbt_node, ns['resqml2'] + 'Geometry')
        geom.set(ns['xsi'] + 'type', ns['resqml2'] + 'ParametricLineGeometry')
        geom.text = '\n'

        # note: resqml standard allows trajectory to be in different crs to md datum
        #       however, this module often uses the md datum crs, if the trajectory has been imported

        kc_node = rqet.SubElement(geom, ns['resqml2'] + 'KnotCount')
        kc_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
        kc_node.text = str(self.knot_count)

        lki_node = rqet.SubElement(geom, ns['resqml2'] + 'LineKindIndex')
        lki_node.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
        lki_node.text = str(self.line_kind_index)

        cpp_node = rqet.SubElement(geom, ns['resqml2'] + 'ControlPointParameters')
        cpp_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'DoubleHdf5Array')
        cpp_node.text = rqet.null_xml_text

        cpp_values_node = rqet.SubElement(cpp_node, ns['resqml2'] + 'Values')
        cpp_values_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        cpp_values_node.text = rqet.null_xml_text

        cp_node = rqet.SubElement(geom, ns['resqml2'] + 'ControlPoints')
        cp_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
        cp_node.text = rqet.null_xml_text

        cp_coords_node = rqet.SubElement(cp_node, ns['resqml2'] + 'Coordinates')
        cp_coords_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
        cp_coords_node.text = rqet.null_xml_text

        if self.tangent_vectors is not None:
            tv_node = rqet.SubElement(geom, ns['resqml2'] + 'TangentVectors')
            tv_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Point3dHdf5Array')
            tv_node.text = rqet.null_xml_text

            tv_coords_node = rqet.SubElement(tv_node, ns['resqml2'] + 'Coordinates')
            tv_coords_node.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
            tv_coords_node.text = rqet.null_xml_text
        else:
            tv_node = None
            tv_coords_node = None

        return geom, kc_node, lki_node, cpp_node, cpp_values_node, cp_node, cp_coords_node, tv_node, tv_coords_node

    def __get_crs_uuid(self):
        """ Assign the same crs uuid to the Trajectory object.

        note: this assumes that the crs is the same for the MdDatum object and the Trajectory object.
        """

        if self.crs_uuid is None:
            self.crs_uuid = self.md_datum.crs_uuid
        assert self.crs_uuid is not None

    def __create_deviation_survey_reference_node(self, wbt_node):
        """ Create a reference node to a DeviationSurvey object and append it to the WellboreTrajectory

        object's root node.
        """

        if self.deviation_survey is not None:
            ds_root = self.deviation_survey.root_node
            self.model.create_ref_node('DeviationSurvey',
                                       rqet.find_tag(rqet.find_tag(ds_root, 'Citation'), 'Title').text,
                                       bu.uuid_from_string(ds_root.attrib['uuid']),
                                       content_type = 'obj_DeviationSurveyRepresentation',
                                       root = wbt_node)

    def __create_wellbore_interpretation_reference_node(self, wbt_node):
        """Create a reference node to a WellboreInterpretation object and append it to the WellboreTrajectory

        object's root node.
        """

        interp_root = None
        if self.wellbore_interpretation is not None:
            interp_root = self.wellbore_interpretation.root
            self.model.create_ref_node('RepresentedInterpretation',
                                       rqet.find_nested_tags_text(interp_root, ['Citation', 'Title']),
                                       bu.uuid_from_string(interp_root.attrib['uuid']),
                                       content_type = 'obj_WellboreInterpretation',
                                       root = wbt_node)
        return interp_root

    def __add_as_part_and_add_relationships(self, wbt_node, interp_root, add_as_part, add_relationships, ext_uuid):
        """Add the newly created Trajectory object's root node as a part in the model and add reciprocal

        relationships.
        """

        if add_as_part:
            self.model.add_part('obj_WellboreTrajectoryRepresentation', self.uuid, wbt_node)
            if add_relationships:
                crs_root = self.model.root_for_uuid(self.crs_uuid)
                self.model.create_reciprocal_relationship(wbt_node, 'destinationObject', crs_root, 'sourceObject')
                self.model.create_reciprocal_relationship(wbt_node, 'destinationObject', self.md_datum.root,
                                                          'sourceObject')
                if self.deviation_survey is not None:
                    self.model.create_reciprocal_relationship(wbt_node, 'destinationObject',
                                                              self.deviation_survey.root_node, 'sourceObject')
                if interp_root is not None:
                    self.model.create_reciprocal_relationship(wbt_node, 'destinationObject', interp_root,
                                                              'sourceObject')
                ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
                ext_node = self.model.root_for_part(ext_part)
                self.model.create_reciprocal_relationship(wbt_node, 'mlToExternalPartProxy', ext_node,
                                                          'externalPartProxyToMl')

    def write_hdf5(self, file_name = None, mode = 'a'):
        """Create or append to an hdf5 file, writing datasets for the measured depths, control points and tangent

        vectors.

        :meta common:
        """

        # NB: array data must all have been set up prior to calling this function
        if self.uuid is None:
            self.uuid = bu.new_uuid()

        h5_reg = rwh5.H5Register(self.model)
        h5_reg.register_dataset(self.uuid, 'controlPointParameters', self.measured_depths)
        h5_reg.register_dataset(self.uuid, 'controlPoints', self.control_points)
        if self.tangent_vectors is not None:
            h5_reg.register_dataset(self.uuid, 'tangentVectors', self.tangent_vectors)
        h5_reg.write(file = file_name, mode = mode)

    def __eq__(self, other):
        """Implements equals operator.

        Compares class type and uuid
        """

        # TODO: more detailed equality comparison
        other_uuid = getattr(other, "uuid", None)
        return isinstance(other, self.__class__) and bu.matching_uuids(self.uuid, other_uuid)



"""Stratigraphy related classes and valid values."""

__all__ = [
    'BinaryContactInterpretation', 'GeologicUnitInterpretation', 'StratigraphicColumnRank', 'StratigraphicColumn',
    'StratigraphicUnitFeature', 'StratigraphicUnitInterpretation', 'valid_compositions', 'valid_implacements',
    'valid_domains', 'valid_deposition_modes', 'valid_ordering_criteria', 'valid_contact_relationships',
    'valid_contact_verbs', 'valid_contact_sides', 'valid_contact_modes'
]

from ._strata_common import valid_compositions, valid_implacements, valid_domains, valid_deposition_modes,  \
    valid_ordering_criteria, valid_contact_relationships, valid_contact_verbs, valid_contact_sides,  \
    valid_contact_modes
from ._binary_contact_interpretation import BinaryContactInterpretation
from ._geologic_unit_interpretation import GeologicUnitInterpretation
from ._stratigraphic_column_rank import StratigraphicColumnRank
from ._stratigraphic_column import StratigraphicColumn
from ._stratigraphic_unit_feature import StratigraphicUnitFeature
from ._stratigraphic_unit_interpretation import StratigraphicUnitInterpretation

# Set "module" attribute of all public objects to this path. Fixes issue #310
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Class for RESQML Fault Interpretation organizational objects."""

import math as maths

import resqpy.olio.uuid as bu
import resqpy.olio.xml_et as rqet
import resqpy.organize
import resqpy.organize.tectonic_boundary_feature as tbf
import resqpy.organize._utils as ou
from resqpy.olio.base import BaseResqpy
from resqpy.olio.xml_namespaces import curly_namespace as ns


class FaultInterpretation(BaseResqpy):
    """Class for RESQML Fault Interpretation organizational objects.

    RESQML documentation:

       A type of boundary feature, this class contains the data describing an opinion
       about the characterization of the fault, which includes the attributes listed below.
    """

    resqml_type = "FaultInterpretation"
    valid_domains = ('depth', 'time', 'mixed')

    # note: many of the attributes could be deduced from geometry

    def __init__(self,
                 parent_model,
                 uuid = None,
                 title = None,
                 tectonic_boundary_feature = None,
                 domain = 'depth',
                 is_normal = None,
                 is_listric = None,
                 maximum_throw = None,
                 mean_azimuth = None,
                 mean_dip = None,
                 originator = None,
                 extra_metadata = None):
        """Initialises a Fault interpretation organisational object."""

        # note: will create a paired TectonicBoundaryFeature object when loading from xml
        # if not extracting from xml,:
        # tectonic_boundary_feature is required and must be a TectonicBoundaryFeature object
        # domain is required and must be one of 'depth', 'time' or 'mixed'
        # is_listric is required if the fault is not normal (and must be None if normal)
        # max throw, azimuth & dip are all optional
        # the throw interpretation list is not supported for direct initialisation

        self.tectonic_boundary_feature = tectonic_boundary_feature  # InterpretedFeature RESQML field, when not loading from xml
        self.feature_root = None if self.tectonic_boundary_feature is None else self.tectonic_boundary_feature.root
        if (not title) and self.tectonic_boundary_feature is not None:
            title = self.tectonic_boundary_feature.feature_name
        self.main_has_occurred_during = (None, None)
        self.is_normal = is_normal  # extra field, not explicitly in RESQML
        self.domain = domain
        # RESQML xml business rule: IsListric must be present if the fault is normal; must not be present if the fault is not normal
        self.is_listric = is_listric
        self.maximum_throw = maximum_throw
        self.mean_azimuth = mean_azimuth
        self.mean_dip = mean_dip
        self.throw_interpretation_list = None  # list of (list of throw kind, (base chrono uuid, top chrono uuid)))

        super().__init__(model = parent_model,
                         uuid = uuid,
                         title = title,
                         originator = originator,
                         extra_metadata = extra_metadata)

    @property
    def feature_uuid(self):
        """Returns the UUID of the interpreted feature"""
        # TODO: rewrite using uuid as primary key
        return rqet.uuid_for_part_root(self.feature_root)

    def _load_from_xml(self):
        root_node = self.root
        self.domain = rqet.find_tag_text(root_node, 'Domain')
        interp_feature_ref_node = rqet.find_tag(root_node, 'InterpretedFeature')
        assert interp_feature_ref_node is not None
        self.feature_root = self.model.referenced_node(interp_feature_ref_node)
        if self.feature_root is not None:
            self.tectonic_boundary_feature = tbf.TectonicBoundaryFeature(self.model,
                                                                         uuid = self.feature_root.attrib['uuid'],
                                                                         feature_name = self.model.title_for_root(
                                                                             self.feature_root))
            self.main_has_occurred_during = ou.extract_has_occurred_during(root_node)
            self.is_listric = rqet.find_tag_bool(root_node, 'IsListric')
            self.is_normal = (self.is_listric is None)
            self.maximum_throw = rqet.find_tag_float(root_node, 'MaximumThrow')
            # todo: check that type="eml:LengthMeasure" is simple float
            self.mean_azimuth = rqet.find_tag_float(root_node, 'MeanAzimuth')
            self.mean_dip = rqet.find_tag_float(root_node, 'MeanDip')
            throw_interpretation_nodes = rqet.list_of_tag(root_node, 'ThrowInterpretation')
            if throw_interpretation_nodes is not None and len(throw_interpretation_nodes):
                self.throw_interpretation_list = []
                for ti_node in throw_interpretation_nodes:
                    hod_pair = ou.extract_has_occurred_during(ti_node)
                    throw_kind_list = rqet.list_of_tag(ti_node, 'Throw')
                    for tk_node in throw_kind_list:
                        self.throw_interpretation_list.append((tk_node.text, hod_pair))

    def is_equivalent(self, other, check_extra_metadata = True):
        """Returns True if this interpretation is essentially the same as the other; otherwise False."""

        if other is None or not isinstance(other, FaultInterpretation):
            return False
        if self is other or bu.matching_uuids(self.uuid, other.uuid):
            return True
        if self.title != other.title:
            return False

        attr_list = ['tectonic_boundary_feature', 'maximum_throw', 'mean_azimuth', 'mean_dip']
        one_none_attr_list = [(getattr(self, v) is None) != (getattr(other, v) is None) for v in attr_list]
        if any(one_none_attr_list):
            # If only one of self or other has a None attribute
            return False

        # List of attributes that are not None in either self or other
        non_none_attr_list = [a for a in attr_list if getattr(self, a) is not None]

        # Additional tests for attributes that are not None
        if (self.tectonic_boundary_feature is not None and
                not self.tectonic_boundary_feature.is_equivalent(other.tectonic_boundary_feature,
                                                                 check_extra_metadata = check_extra_metadata)):
            return False
        if (self.maximum_throw is not None and
                not maths.isclose(self.maximum_throw, other.maximum_throw, rel_tol = 1e-3)):
            return False
        if (self.mean_azimuth is not None and not maths.isclose(self.mean_azimuth, other.mean_azimuth, abs_tol = 0.5)):
            return False
        if (self.mean_dip is not None and not maths.isclose(self.mean_dip, other.mean_dip, abs_tol = 0.5)):
            return False

        if (not ou.equivalent_chrono_pairs(self.main_has_occurred_during, other.main_has_occurred_during) or
                self.is_normal != other.is_normal or self.domain != other.domain or
                self.is_listric != other.is_listric):
            return False

        if check_extra_metadata and not ou.equivalent_extra_metadata(self, other):
            return False
        if not self.throw_interpretation_list and not other.throw_interpretation_list:
            return True
        if not self.throw_interpretation_list or not other.throw_interpretation_list:
            return False
        if len(self.throw_interpretation_list) != len(other.throw_interpretation_list):
            return False
        for this_ti, other_ti in zip(self.throw_interpretation_list, other.throw_interpretation_list):
            if this_ti[0] != other_ti[0]:
                return False  # throw kind
            if not ou.equivalent_chrono_pairs(this_ti[1], other_ti[1]):
                return False
        return True

    def create_xml(self,
                   tectonic_boundary_feature_root = None,
                   add_as_part = True,
                   add_relationships = True,
                   originator = None,
                   title_suffix = None,
                   reuse = True):
        """Creates a fault interpretation organisational xml node from a fault interpretation object."""

        # note: related tectonic boundary feature node should be created first and referenced here

        assert self.is_normal == (self.is_listric is None)
        if not self.title:
            if tectonic_boundary_feature_root is not None:
                title = rqet.find_nested_tags_text(tectonic_boundary_feature_root, ['Citation', 'Title'])
            else:
                title = 'fault interpretation'
            if title_suffix:
                title += ' ' + title_suffix

        if reuse and self.try_reuse():
            return self.root

        fi = super().create_xml(add_as_part = False, originator = originator)

        if self.tectonic_boundary_feature is not None:
            tbf_root = self.tectonic_boundary_feature.root
            if tbf_root is not None:
                if tectonic_boundary_feature_root is None:
                    tectonic_boundary_feature_root = tbf_root
                else:
                    assert tbf_root is tectonic_boundary_feature_root, 'tectonic boundary feature mismatch'
        assert tectonic_boundary_feature_root is not None

        assert self.domain in self.valid_domains, 'illegal domain value for fault interpretation'
        dom_node = rqet.SubElement(fi, ns['resqml2'] + 'Domain')
        dom_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'Domain')
        dom_node.text = self.domain

        self.model.create_ref_node('InterpretedFeature',
                                   self.model.title_for_root(tectonic_boundary_feature_root),
                                   tectonic_boundary_feature_root.attrib['uuid'],
                                   content_type = 'obj_TectonicBoundaryFeature',
                                   root = fi)

        ou.create_xml_has_occurred_during(self.model, fi, self.main_has_occurred_during)

        if self.is_listric is not None:
            listric = rqet.SubElement(fi, ns['resqml2'] + 'IsListric')
            listric.set(ns['xsi'] + 'type', ns['xsd'] + 'boolean')
            listric.text = str(self.is_listric).lower()

        # todo: check eml:LengthMeasure and PlaneAngleMeasure structures: uom?
        if self.maximum_throw is not None:
            max_throw = rqet.SubElement(fi, ns['resqml2'] + 'MaximumThrow')
            max_throw.set(ns['xsi'] + 'type', ns['eml'] + 'LengthMeasure')
            max_throw.text = str(self.maximum_throw)

        if self.mean_azimuth is not None:
            azimuth = rqet.SubElement(fi, ns['resqml2'] + 'MeanAzimuth')
            azimuth.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
            azimuth.text = str(self.mean_azimuth)

        if self.mean_dip is not None:
            dip = rqet.SubElement(fi, ns['resqml2'] + 'MeanDip')
            dip.set(ns['xsi'] + 'type', ns['eml'] + 'PlaneAngleMeasure')
            dip.text = str(self.mean_azimuth)

        if self.throw_interpretation_list is not None and len(self.throw_interpretation_list):
            previous_has_occurred_during = ('never', 'never')
            ti_node = None
            for (throw_kind, (base_chrono_uuid, top_chrono_uuid)) in self.throw_interpretation_list:
                if (str(base_chrono_uuid), str(top_chrono_uuid)) != previous_has_occurred_during:
                    previous_has_occurred_during = (str(base_chrono_uuid), str(top_chrono_uuid))
                    ti_node = rqet.SubElement(fi, 'ThrowInterpretation')
                    ti_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'FaultThrow')
                    ti_node.text = rqet.null_xml_text
                    ou.create_xml_has_occurred_during(self.model, ti_node, (base_chrono_uuid, top_chrono_uuid))
                tk_node = rqet.SubElement(ti_node, 'Throw')
                tk_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ThrowKind')
                tk_node.text = throw_kind

        if add_as_part:
            self.model.add_part('obj_FaultInterpretation', self.uuid, fi)
            if add_relationships:
                self.model.create_reciprocal_relationship(fi, 'destinationObject', tectonic_boundary_feature_root,
                                                          'sourceObject')

        return fi



"""High level add faults function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np

import resqpy.crs as rqc
import resqpy.derived_model
import resqpy.grid as grr
import resqpy.lines as rql
import resqpy.model as rq
import resqpy.olio.grid_functions as gf
import resqpy.olio.simple_lines as sl
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.derived_model._common as rqdm_c
import resqpy.derived_model._copy_grid as rqdm_cg


def add_faults(epc_file,
               source_grid,
               polylines = None,
               lines_file_list = None,
               lines_crs_uuid = None,
               full_pillar_list_dict = None,
               left_right_throw_dict = None,
               create_gcs = True,
               inherit_properties = False,
               inherit_realization = None,
               inherit_all_realizations = False,
               new_grid_title = None,
               new_epc_file = None):
    """Extends epc file with a new grid which is a version of the source grid with new curtain fault(s) added.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       polylines (lines.PolylineSet or list of lines.Polyline, optional): list of poly lines for which curtain faults
          are to be added; either this or lines_file_list or full_pillar_list_dict must be present
       lines_file_list (list of str, optional): a list of file paths, each containing one or more poly lines in simple
          ascii format§; see notes; either this or polylines or full_pillar_list_dicr must be present
       lines_crs_uuid (uuid, optional): if present, the uuid of a coordinate reference system with which to interpret
          the contents of the lines files; if None, the crs used by the grid will be assumed
       full_pillar_list_dict (dict mapping str to list of pairs of ints, optional): dictionary mapping from a fault name
          to a list of pairs of ints being the ordered neigbouring primary pillar (j0, i0) defining the curtain fault;
          either this or polylines or lines_file_list must be present
       left_right_throw_dict (dict mapping str to pair of floats, optional): dictionary mapping from a fault name to a
          pair of floats being the semi-throw adjustment on the left and the right of the fault (see notes); semi-throw
          values default to (+0.5, -0.5)
       create_gcs (boolean, default True): if True, and faults are being defined by lines, a grid connection set is
          created with one feature per new fault and associated organisational objects are also created; ignored if
          lines_file_list is None
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the unsplit grid (& crs)

    returns:
       a new grid (grid.Grid object) which is a copy of the source grid with the structure modified to incorporate
       the new faults

    notes:
       full_pillar_list_dict is typically generated by Grid.make_face_sets_from_pillar_lists();
       pillars will be split as needed to model the new faults, though existing splits will be used as appropriate, so
       this function may also be used to add a constant to the throw of existing faults;
       the left_right_throw_dict contains a pair of floats for each fault name (as found in keys of full_pillar_list_dict);
       these throw values are lengths in the uom of the crs used by the grid (which must have the same xy units as z units);

       this function does not add a GridConnectionSet to the model – calling code may wish to do that
    """

    log.info('adding faults')

    assert epc_file or new_epc_file, 'epc file name not specified'
    assert epc_file or source_grid is not None, 'neither epc file name nor source grid supplied'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = rqdm_c._establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation in ['IjkGrid', 'IjkBlockGrid']  # unstructured grids not catered for
    assert model is not None
    assert len([arg for arg in (polylines, lines_file_list, full_pillar_list_dict) if arg is not None]) == 1

    # take a copy of the resqpy grid object, without writing to hdf5 or creating xml
    # the copy will be a Grid, even if the source is a RegularGrid
    grid = rqdm_cg.copy_grid(source_grid, model)
    grid.crs_uuid = source_grid.crs_uuid
    if source_grid.model is not model:
        model.duplicate_node(source_grid.model.root_for_uuid(grid.crs_uuid), add_as_part = True)
    grid.crs = rqc.Crs(model, uuid = grid.crs_uuid)

    if isinstance(polylines, rql.PolylineSet):
        polylines = polylines.convert_to_polylines()

    composite_face_set_dict = {}

    # build pillar list dict for polylines if necessary
    if full_pillar_list_dict is None:
        full_pillar_list_dict = {}
        _populate_composite_face_sets_for_polylines(model, grid, polylines, lines_crs_uuid, grid.crs, lines_file_list,
                                                    full_pillar_list_dict, composite_face_set_dict)

    else:  # populate composite face set dictionary from full pillar list
        _populate_composite_face_sets_for_pillar_lists(source_grid, full_pillar_list_dict, composite_face_set_dict)

    # log.debug(f'full_pillar_list_dict:\n{full_pillar_list_dict}')

    _process_full_pillar_list_dict(grid, full_pillar_list_dict, left_right_throw_dict)

    collection = rqdm_c._prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization,
                                                    inherit_all_realizations)
    # todo: recompute depth properties (and volumes, cell lengths etc. if being strict)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'copy of ' + str(rqet.citation_title_for_node(source_grid.root)) + ' with added faults'

    # write model
    if new_epc_file:
        rqdm_c._write_grid(new_epc_file,
                           grid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'w')
    else:
        ext_uuid = model.h5_uuid()

        rqdm_c._write_grid(epc_file,
                           grid,
                           ext_uuid = ext_uuid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'a')

    # create grid connection set if requested
    _create_gcs_if_requested(create_gcs, composite_face_set_dict, new_epc_file, grid)

    return grid


def _make_face_sets_for_new_lines(new_lines, face_set_id, grid, full_pillar_list_dict, composite_face_set_dict):
    """Adds entries to full_pillar_list_dict & composite_face_set_dict for new lines."""
    pillar_list_list = sl.nearest_pillars(new_lines, grid)
    face_set_dict, full_pll_dict = grid.make_face_sets_from_pillar_lists(pillar_list_list, face_set_id)
    for key, pll in full_pll_dict.items():
        full_pillar_list_dict[key] = pll
    for key, fs_info in face_set_dict.items():
        composite_face_set_dict[key] = fs_info


def _populate_composite_face_sets_for_pillar_lists(grid, full_pillar_list_dict, composite_face_set_dict):
    for key, pillar_list in full_pillar_list_dict.items():
        face_set_dict, _ = grid.make_face_sets_from_pillar_lists([pillar_list], key)
        for k, fs_info in face_set_dict.items():
            composite_face_set_dict[k] = fs_info


def _fault_from_pillar_list(grid, full_pillar_list, delta_throw_left, delta_throw_right):
    """Creates and/or adjusts throw on a single fault defined by a full pillar list, in memory.

    arguments:
       grid (grid.Grid): the grid object to be adjusted in memory (should have originally been copied
          without the hdf5 arrays having been written yet, nor xml created)
       full_pillar_list (list of pairs of ints (j0, i0)): the full list of primary pillars defining
          the fault; neighbouring pairs must differ by exactly one in either j0 or i0 but not both
       delta_throw_left (float): the amount to add to the 'depth' of points to the left of the line
          when viewed from above, looking along the line in the direction of the pillar list entries;
          units are implicitly the length units of the crs used by the grid; see notes about 'depth'
       delta_throw_right (float): as for delta_throw_left but applied to points to the right of the
          line
    """

    # this function introduces new data into the RESQML arrays representing split pillars
    # familiarity with those array representations is needed if working on this function

    if full_pillar_list is None or len(full_pillar_list) < 3:
        return
    assert grid.z_units() == grid.xy_units(),  \
        'crs used by grid has differing xy and z units – mix not supported when adding faults'
    grid.cache_all_geometry_arrays()
    assert hasattr(grid, 'points_cached')
    # make grid into a faulted grid if hitherto unfaulted
    if not grid.has_split_coordinate_lines:
        grid.points_cached = grid.points_cached.reshape((grid.nk_plus_k_gaps + 1, (grid.nj + 1) * (grid.ni + 1), 3))
        grid.split_pillar_indices_cached = np.array([], dtype = int)
        grid.cols_for_split_pillars = np.array([], dtype = int)
        grid.cols_for_split_pillars_cl = np.array([], dtype = int)
        grid.has_split_coordinate_lines = True
    assert grid.points_cached.ndim == 3
    if len(grid.cols_for_split_pillars_cl) == 0:
        cl = 0
    else:
        cl = grid.cols_for_split_pillars_cl[-1]
    original_p = np.zeros((grid.nk_plus_k_gaps + 1, 3), dtype = float)
    n_primaries = (grid.nj + 1) * (grid.ni + 1)
    for p_index in range(1, len(full_pillar_list) - 1):
        primary_ji0 = full_pillar_list[p_index]
        primary = primary_ji0[0] * (grid.ni + 1) + primary_ji0[1]
        p_vector = np.array(_pillar_vector(grid, primary), dtype = float)
        if p_vector is None:
            continue
        throw_left_vector = np.expand_dims(delta_throw_left * p_vector, axis = 0)
        throw_right_vector = np.expand_dims(delta_throw_right * p_vector, axis = 0)
        # log.debug(f'T: p ji0: {primary_ji0}; p vec: {p_vector}; left v: {throw_left_vector}; right v: {throw_right_vector}')
        existing_foursome = grid.pillar_foursome(primary_ji0, none_if_unsplit = False)
        lr_foursome = gf.left_right_foursome(full_pillar_list, p_index)
        cl = _processs_foursome(grid, n_primaries, primary, original_p, existing_foursome, lr_foursome, primary_ji0,
                                throw_right_vector, throw_left_vector, cl)


def _pillar_vector(grid, p_index):
    # return a unit vector for direction of pillar, in direction of increasing k
    if np.all(np.isnan(grid.points_cached[:, p_index])):
        return None
    k_top = 0
    while np.any(np.isnan(grid.points_cached[k_top, p_index])):
        k_top += 1
    k_bot = grid.nk_plus_k_gaps - 1
    while np.any(np.isnan(grid.points_cached[k_bot, p_index])):
        k_bot -= 1
    if k_bot == k_top:  # following coded to treat None directions as downwards
        if grid.k_direction_is_down is False:
            if grid.z_inc_down() is False:
                return (0.0, 0.0, 1.0)
            else:
                return (0.0, 0.0, -1.0)
        else:
            if grid.z_inc_down() is False:
                return (0.0, 0.0, -1.0)
            else:
                return (0.0, 0.0, 1.0)
    else:
        return vec.unit_vector(grid.points_cached[k_bot, p_index] - grid.points_cached[k_top, p_index])


def _extend_points_cached(grid, exist_p):
    s = grid.points_cached.shape
    e = np.empty((s[0], s[1] + 1, s[2]), dtype = float)
    e[:, :-1, :] = grid.points_cached
    e[:, -1, :] = grid.points_cached[:, exist_p, :]
    grid.points_cached = e


def _np_int_extended(a, i):
    e = np.empty(a.size + 1, dtype = int)
    e[:-1] = a
    e[-1] = i
    return e


def _create_gcs_if_requested(create_gcs, composite_face_set_dict, new_epc_file, grid):
    if create_gcs and len(composite_face_set_dict) > 0:
        if new_epc_file is not None:
            grid_uuid = grid.uuid
            model = rq.Model(new_epc_file)
            grid = grr.Grid(model, root = model.root(uuid = grid_uuid), find_properties = False)
        grid.set_face_set_gcs_list_from_dict(composite_face_set_dict, create_organizing_objects_where_needed = True)
        combined_gcs = grid.face_set_gcs_list[0]
        for gcs in grid.face_set_gcs_list[1:]:
            combined_gcs.append(gcs)
        combined_gcs.write_hdf5()
        combined_gcs.create_xml(title = 'faults added from lines')
        grid.clear_face_sets()
        grid.model.store_epc()


def _processs_foursome(grid, n_primaries, primary, original_p, existing_foursome, lr_foursome, primary_ji0,
                       throw_right_vector, throw_left_vector, cl):
    p_j, p_i = primary_ji0
    # log.debug(f'P: p ji0: {primary_ji0}; e foursome:\n{existing_foursome}; lr foursome:\n{lr_foursome}')
    for exist_p in np.unique(existing_foursome):
        exist_lr = None
        new_p_made = False
        for jp in range(2):
            if (p_j == 0 and jp == 0) or (p_j == grid.nj and jp == 1):
                continue
            for ip in range(2):
                if (p_i == 0 and ip == 0) or (p_i == grid.ni and ip == 1):
                    continue
                if existing_foursome[jp, ip] != exist_p:
                    continue
                if exist_lr is None:
                    original_p[:] = grid.points_cached[:, exist_p, :]
                    exist_lr = lr_foursome[jp, ip]
                    # log.debug(f'A: p ji0: {primary_ji0}; exist_p: {exist_p}; jp,ip: {(jp,ip)}; exist_lr: {exist_lr}')
                    grid.points_cached[:, exist_p, :] += throw_right_vector if exist_lr else throw_left_vector
                    continue
                if lr_foursome[jp, ip] == exist_lr:
                    continue
                natural_col = (p_j + jp - 1) * grid.ni + p_i + ip - 1
                if exist_p != primary:  # remove one of the columns currently assigned to exist_p
                    extra_p = exist_p - n_primaries
                    # log.debug(f're-split: primary: {primary}; exist: {exist_p}; col: {natural_col}; extra: {extra_p}')
                    # log.debug(f'pre re-split: cols: {grid.cols_for_split_pillars}')
                    # log.debug(f'pre re-split: ccl:  {grid.cols_for_split_pillars_cl}')
                    assert grid.split_pillar_indices_cached[extra_p] == primary
                    if extra_p == 0:
                        start = 0
                    else:
                        start = grid.cols_for_split_pillars_cl[extra_p - 1]
                    found = False
                    for cols_index in range(start, start + grid.cols_for_split_pillars_cl[extra_p]):
                        if grid.cols_for_split_pillars[cols_index] == natural_col:
                            grid.cols_for_split_pillars = np.concatenate((grid.cols_for_split_pillars[:cols_index],
                                                                          grid.cols_for_split_pillars[cols_index + 1:]))
                            found = True
                            break
                    assert found
                    grid.cols_for_split_pillars_cl[extra_p:] -= 1
                    cl -= 1
                    assert grid.cols_for_split_pillars_cl[extra_p] > 0
                # log.debug(f'post re-split: cols: {grid.cols_for_split_pillars}')
                # log.debug(f'post re-split: ccl:  {grid.cols_for_split_pillars_cl}')
                if not new_p_made:  # create a new split of pillar
                    _extend_points_cached(grid, exist_p)
                    # log.debug(f'B: p ji0: {primary_ji0}; exist_p: {exist_p}; jp,ip: {(jp,ip)}; lr: {lr_foursome[jp, ip]}; c ji0: {natural_col}')
                    grid.points_cached[:, -1, :] = original_p + (throw_right_vector
                                                                 if lr_foursome[jp, ip] else throw_left_vector)
                    grid.split_pillar_indices_cached = _np_int_extended(grid.split_pillar_indices_cached, primary)
                    if grid.split_pillars_count is None:
                        grid.split_pillars_count = 0
                    grid.split_pillars_count += 1
                    grid.cols_for_split_pillars = _np_int_extended(grid.cols_for_split_pillars, natural_col)
                    cl += 1
                    grid.cols_for_split_pillars_cl = _np_int_extended(grid.cols_for_split_pillars_cl, cl)
                    new_p_made = True
                else:  # include this column in newly split version of pillar
                    # log.debug(f'C: p ji0: {primary_ji0}; exist_p: {exist_p}; jp,ip: {(jp,ip)}; lr: {lr_foursome[jp, ip]}; c ji0: {natural_col}')
                    grid.cols_for_split_pillars = _np_int_extended(grid.cols_for_split_pillars, natural_col)
                    cl += 1
                    grid.cols_for_split_pillars_cl[-1] = cl
    return cl


def _process_full_pillar_list_dict(grid, full_pillar_list_dict, left_right_throw_dict):
    for fault_key in full_pillar_list_dict:
        full_pillar_list = full_pillar_list_dict[fault_key]
        left_right_throw = None
        if left_right_throw_dict is not None:
            left_right_throw = left_right_throw_dict.get(fault_key)
        if left_right_throw is None:
            left_right_throw = (+0.5, -0.5)
        log.debug(
            f'generating fault {fault_key} pillar count {len(full_pillar_list)}; left, right throw {left_right_throw}')
        _fault_from_pillar_list(grid, full_pillar_list, left_right_throw[0], left_right_throw[1])


def _populate_composite_face_sets_for_polylines(model, grid, polylines, lines_crs_uuid, grid_crs, lines_file_list,
                                                full_pillar_list_dict, composite_face_set_dict):
    lines_crs = None if lines_crs_uuid is None else rqc.Crs(model, uuid = lines_crs_uuid)
    if polylines:
        for i, polyline in enumerate(polylines):
            new_line = polyline.coordinates.copy()
            if polyline.crs_uuid is not None and polyline.crs_uuid != lines_crs_uuid:
                lines_crs_uuid = polyline.crs_uuid
                lines_crs = rqc.Crs(model, uuid = lines_crs_uuid)
            if lines_crs:
                lines_crs.convert_array_to(grid_crs, new_line)
            title = polyline.title if polyline.title else 'fault_' + str(i)
            _make_face_sets_for_new_lines([new_line], title, grid, full_pillar_list_dict, composite_face_set_dict)
    else:
        for filename in lines_file_list:
            new_lines = sl.read_lines(filename)
            if lines_crs is not None:
                for a in new_lines:
                    lines_crs.convert_array_to(grid_crs, a)
            _, f_name = os.path.split(filename)
            if f_name.lower().endswith('.dat'):
                face_set_id = f_name[:-4]
            else:
                face_set_id = f_name
            _make_face_sets_for_new_lines(new_lines, face_set_id, grid, full_pillar_list_dict, composite_face_set_dict)



"""The Grid Module."""

__all__ = [
    'Grid', 'RegularGrid', 'extract_grid_parent', 'establish_zone_property_kind', 'find_cell_for_x_sect_xz',
    'grid_flavour', 'is_regular_grid', 'any_grid'
]

from ._grid import Grid
from ._regular_grid import RegularGrid
from ._grid_types import grid_flavour, is_regular_grid, any_grid
from ._moved_functions import establish_zone_property_kind
from ._extract_functions import extract_grid_parent, extent_kji_from_root
from ._points_functions import find_cell_for_x_sect_xz

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""High level drape to surface function."""

import logging

log = logging.getLogger(__name__)

import os
import numpy as np
import copy

import resqpy.crs as rqc
import resqpy.derived_model
import resqpy.grid_surface as rgs
import resqpy.model as rq
import resqpy.olio.intersection as meet
import resqpy.olio.xml_et as rqet

import resqpy.derived_model._common as rqdm_c
import resqpy.derived_model._copy_grid as rqdm_cg


def drape_to_surface(epc_file,
                     source_grid = None,
                     surface = None,
                     scaling_factor = None,
                     ref_k0 = 0,
                     ref_k_faces = 'top',
                     quad_triangles = True,
                     border = None,
                     store_displacement = False,
                     inherit_properties = False,
                     inherit_realization = None,
                     inherit_all_realizations = False,
                     new_grid_title = None,
                     new_epc_file = None):
    """Return a new grid with geometry draped to a surface.
    
    Extend a resqml model with a new grid where the reference layer boundary of the source grid has been re-draped
    to a surface.

    arguments:
       epc_file (string): file name to rewrite the model's xml to; if source grid is None, model is loaded from this file
       source_grid (grid.Grid object, optional): if None, the epc_file is loaded and it should contain one ijk grid object
          (or one 'ROOT' grid) which is used as the source grid
       surface (surface.Surface object, optional): the surface to drape the grid to; if None, a surface is generated from
          the reference layer boundary (which can then be scaled with the scaling_factor)
       scaling_factor (float, optional): if not None, prior to draping, the surface is stretched vertically by this factor,
          away from a horizontal plane located at the surface's shallowest depth
       ref_k0 (integer, default 0): the reference layer (zero based) to drape to the surface
       ref_k_faces (string, default 'top'): 'top' or 'base' identifying which bounding interface to use as the reference
       quad_triangles (boolean, default True): if True and surface is None, each cell face in the reference boundary layer
          is represented by 4 triangles (with a common vertex at the face centre) in the generated surface; if False,
          only 2 trianges are used for each cell face (which gives a non-unique solution)
       cell_range (integer, default 0): the number of cells away from faults which will have depths adjusted to spatially
          smooth the effect of the throw scaling (ie. reduce sudden changes in gradient due to the scaling)
       offset_decay (float, default 0.5): the factor to reduce depth shifts by with each cell step away from faults (used
          in conjunction with cell_range)
       store_displacement (boolean, default False): if True, 3 grid property parts are created, one each for x, y, & z
          displacement of cells' centres brought about by the local depth shift
       inherit_properties (boolean, default False): if True, the new grid will have a copy of any properties associated
          with the source grid
       inherit_realization (int, optional): realization number for which properties will be inherited; ignored if
          inherit_properties is False
       inherit_all_realizations (boolean, default False): if True (and inherit_realization is None), properties for all
          realizations will be inherited; if False, only properties with a realization of None are inherited; ignored if
          inherit_properties is False or inherit_realization is not None
       new_grid_title (string): used as the citation title text for the new grid object
       new_epc_file (string, optional): if None, the source epc_file is extended with the new grid object; if present,
          a new epc file (& associated h5 file) is created to contain the draped grid (& crs)

    returns:
       new grid (grid.Grid object), with geometry draped to surface

    notes:
       at least one of a surface or a scaling factor must be given;
       if no surface is given, one is created from the fault-healed grid points for the reference layer interface;
       if a scaling factor other than 1.0 is given, the surface is flexed vertically, relative to its shallowest point;
       layer thicknesses measured along pillars are maintained; cell volumes may change;
       the coordinate reference systems for the surface and the grid are assumed to be the same;
       this function currently uses an exhaustive, computationally and memory intensive algorithm;
       setting quad_triangles argument to False should give a factor of 2 speed up and reduction in memory requirement;
       the epc file and associated hdf5 file are appended to (extended) with the new grid, as a side effect of this function
    """

    log.info('draping grid to surface')

    assert epc_file or new_epc_file, 'epc file name not specified'
    if new_epc_file and epc_file and (
        (new_epc_file == epc_file) or
        (os.path.exists(new_epc_file) and os.path.exists(epc_file) and os.path.samefile(new_epc_file, epc_file))):
        new_epc_file = None
    model, source_grid = rqdm_c._establish_model_and_source_grid(epc_file, source_grid)
    assert source_grid.grid_representation == 'IjkGrid'
    assert model is not None

    assert ref_k0 >= 0 and ref_k0 < source_grid.nk
    assert ref_k_faces in ['top', 'base']
    assert surface is not None or (scaling_factor is not None and scaling_factor != 1.0)

    if surface is None:
        surface = rgs.generate_untorn_surface_for_layer_interface(source_grid,
                                                                  k0 = ref_k0,
                                                                  ref_k_faces = ref_k_faces,
                                                                  quad_triangles = quad_triangles,
                                                                  border = border)

    if scaling_factor is not None and scaling_factor != 1.0:
        scaled_surf = copy.deepcopy(surface)
        scaled_surf.vertical_rescale_points(scaling_factor = scaling_factor)
        surface = scaled_surf

    # check that surface and grid use same crs; if not, convert to same
    surface_crs = rqc.Crs(surface.model, surface.crs_uuid)
    if source_grid.crs is None:
        source_grid.crs = rqc.Crs(source_grid.model, uuid = source_grid.crs_uuid)
    if surface_crs != source_grid.crs:
        surface.triangles_and_points()
        surface_crs.convert_array_to(source_grid.crs, surface.points)

    # take a copy of the grid
    log.debug('copying grid')
    grid = rqdm_cg.copy_grid(source_grid, model)
    grid.cache_all_geometry_arrays()  # probably already cached anyway

    # todo: handle pillars with no geometry defined, and cells without geometry defined
    assert grid.geometry_defined_for_all_pillars(), 'not all pillars have defined geometry'
    assert grid.geometry_defined_for_all_cells(), 'not all cells have defined geometry'

    # fetch unsplit equivalent of grid points
    log.debug('fetching unsplit equivalent grid points')
    unsplit_points = grid.unsplit_points_ref(cache_array = True)

    # assume pillars to be straight lines based on top and base points
    log.debug('setting up pillar sample points and directional vectors')
    line_p = unsplit_points[0, :, :, :].reshape((-1, 3))
    line_v = unsplit_points[-1, :, :, :].reshape((-1, 3)) - line_p
    if ref_k_faces == 'base':
        ref_k0 += 1

    # access triangulated surface as triangle node indices into array of points
    log.debug('fetching surface points and triangle corner indices')
    t, p = surface.triangles_and_points()  # will pick up cached crs converted copy if appropriate

    # compute intersections of all pillars with all triangles (sparse array returned with NaN for no intersection)
    log.debug('computing intersections of all pillars with all triangles')
    intersects = meet.line_set_triangles_intersects(line_p, line_v, p[t])

    # reduce to a single intersection point per pillar; todo: flag multiple intersections with a warning
    log.debug('selecting last intersection for each pillar (there should be only one intersection anyway)')
    picks = meet.last_intersects(intersects)

    # count the number of pillars with no intersection at surface (indicated by triple nan)
    log.debug('counting number of pillars which fail to intersect with surface')
    failures = np.count_nonzero(np.isnan(picks)) // 3
    log.info('number of pillars which do not intersect with surface: ' + str(failures))
    assert failures == 0, 'cannot proceed as some pillars do not intersect with surface'

    # compute a translation vector per pillar
    log.debug('computing translation vectors for pillars')
    translate = picks - unsplit_points[ref_k0, :, :, :].reshape((-1, 3))

    # shift all points by translation vectors
    _shift_by_translation_vectors(grid, translate)

    # check cell edge relative directions (in x,y) to ensure geometry is still coherent
    log.debug('checking grid geometry coherence')
    grid.check_top_and_base_cell_edge_directions()

    # build cell displacement property array(s)
    if store_displacement:
        log.debug('generating cell displacement property arrays')
        displacement_collection = rqdm_c._displacement_properties(grid, source_grid)
    else:
        displacement_collection = None

    collection = rqdm_c._prepare_simple_inheritance(grid, source_grid, inherit_properties, inherit_realization,
                                                    inherit_all_realizations)
    if collection is None:
        collection = displacement_collection
    elif displacement_collection is not None:
        collection.inherit_imported_list_from_other_collection(displacement_collection, copy_cached_arrays = False)

    if new_grid_title is None or len(new_grid_title) == 0:
        new_grid_title = 'grid flexed from ' + str(rqet.citation_title_for_node(source_grid.root))

    # write model
    model.h5_release()
    if new_epc_file:
        rqdm_c._write_grid(new_epc_file,
                           grid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'w')
    else:
        ext_uuid, _ = model.h5_uuid_and_path_for_node(rqet.find_nested_tags(source_grid.root, ['Geometry', 'Points']),
                                                      'Coordinates')
        rqdm_c._write_grid(epc_file,
                           grid,
                           ext_uuid = ext_uuid,
                           property_collection = collection,
                           grid_title = new_grid_title,
                           mode = 'a')

    return grid


def _shift_by_translation_vectors(grid, translate):
    log.debug('shifting entire grid along pillars')
    if grid.has_split_coordinate_lines:
        jip1 = (grid.nj + 1) * (grid.ni + 1)
        # adjust primary pillars
        grid.points_cached[:, :jip1, :] += translate.reshape((1, jip1, 3))  # will be broadcast over k axis
        # adjust split pillars
        for p in range(grid.split_pillars_count):
            primary = grid.split_pillar_indices_cached[p]
            grid.points_cached[:, jip1 + p, :] += translate.reshape(
                (1, jip1, 3))[0, primary, :]  # will be broadcast over k axis
    else:
        grid.points_cached[:, :, :, :] +=  \
           translate.reshape((1, grid.points_cached.shape[1], grid.points_cached.shape[2], 3))    # will be broadcast over k axis



"""Submodule containing functions for supporting properties for a property collection."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import resqpy.olio.uuid as bu


def _set_support_model(collection, model, support):
    if model is None and support is not None:
        model = support.model
    if model is None:
        model = collection.model
    else:
        collection.model = model
    return model


def _set_support_uuid_none(collection):
    if collection.support_uuid is not None:
        log.warning('clearing supporting representation for property collection')
    collection.support = None
    collection.support_root = None
    collection.support_uuid = None


def _set_support_uuid_notnone(collection, support, support_uuid, model, modify_parts):
    # when at global level was causing circular reference loading issues as grid imports this module
    import resqpy.fault as rqf
    import resqpy.grid as grr
    import resqpy.surface as rqs
    import resqpy.unstructured as rug
    import resqpy.well as rqw

    assert model is not None, 'model not established when setting support for property collection'
    if collection.support_uuid is not None and not bu.matching_uuids(support_uuid, collection.support_uuid):
        log.warning('changing supporting representation for property collection')
    collection.support_uuid = support_uuid
    collection.support = support
    if collection.support is None:
        _set_support_uuid_notnone_supportnone(collection, support_uuid, model)
    else:
        if type(collection.support) in [
                grr.Grid, grr.RegularGrid, rqw.WellboreFrame, rqw.BlockedWell, rqs.Mesh, rqf.GridConnectionSet,
                rug.UnstructuredGrid, rug.HexaGrid, rug.TetraGrid, rug.PrismGrid, rug.VerticalPrismGrid,
                rug.PyramidGrid, rqw.WellboreMarkerFrame, rqs.Surface
        ]:
            collection.support_root = collection.support.root
        else:
            raise TypeError('unsupported property supporting representation class: ' + str(type(collection.support)))
    if modify_parts:
        for (part, info) in collection.dict.items():
            if info[1] is not None:
                modified = list(info)
                modified[1] = support_uuid
                collection.dict[part] = tuple(modified)


def _set_support_uuid_notnone_supportnone(collection, support_uuid, model):
    import resqpy.fault as rqf
    import resqpy.grid as grr
    import resqpy.surface as rqs
    import resqpy.unstructured as rug
    import resqpy.well as rqw

    support_part = model.part_for_uuid(support_uuid)
    assert support_part is not None, 'supporting representation part missing in model'
    collection.support_root = model.root_for_part(support_part)
    support_type = model.type_of_part(support_part)
    assert support_type is not None
    if support_type == 'obj_IjkGridRepresentation':
        collection.support = grr.any_grid(model, uuid = collection.support_uuid, find_properties = False)
    elif support_type == 'obj_WellboreFrameRepresentation':
        collection.support = rqw.WellboreFrame(model, uuid = collection.support_uuid)
    elif support_type == 'obj_BlockedWellboreRepresentation':
        collection.support = rqw.BlockedWell(model, uuid = collection.support_uuid)
    elif support_type == 'obj_Grid2dRepresentation':
        collection.support = rqs.Mesh(model, uuid = collection.support_uuid)
    elif support_type == 'obj_GridConnectionSetRepresentation':
        collection.support = rqf.GridConnectionSet(model, uuid = collection.support_uuid)
    elif support_type == 'obj_TriangulatedSetRepresentation':
        collection.support = rqs.Surface(model, uuid = collection.support_uuid)
    elif support_type == 'obj_UnstructuredGridRepresentation':
        collection.support = rug.UnstructuredGrid(model,
                                                  uuid = collection.support_uuid,
                                                  geometry_required = False,
                                                  find_properties = False)
    elif support_type == 'obj_WellboreMarkerFrameRepresentation':
        collection.support = rqw.WellboreMarkerFrame(model, uuid = collection.support_uuid)
    else:
        raise TypeError('unsupported property supporting representation class: ' + str(support_type))


def _set_support_and_model_from_collection(collection, other, support_uuid, grid):
    _confirm_support_and_model_from_collection(collection, support_uuid, grid, other)

    assert collection.support_uuid is None or other.support_uuid is None or bu.matching_uuids(
        collection.support_uuid, other.support_uuid)
    if collection.support_uuid is None and collection.number_of_parts() == 0:
        collection.set_support(support_uuid = other.support_uuid, support = other.support)


def _confirm_support_and_model_from_collection(collection, support_uuid, grid, other):
    if support_uuid is None and grid is not None:
        support_uuid = grid.uuid
    if support_uuid is not None and collection.support_uuid is not None:
        assert bu.matching_uuids(support_uuid, collection.support_uuid)

    assert other is not None
    if collection.model is None:
        collection.model = other.model
    else:
        assert collection.model is other.model
